@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2009 Junjiro R. Okajima
+ * Copyright (C) 2005-2011 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -42,7 +42,7 @@ void au_cpup_attr_timesizes(struct inode *inode)
 
 	h_inode = au_h_iptr(inode, au_ibstart(inode));
 	fsstack_copy_attr_times(inode, h_inode);
-	vfsub_copy_inode_size(inode, h_inode);
+	fsstack_copy_inode_size(inode, h_inode);
 }
 
 void au_cpup_attr_nlink(struct inode *inode, int force)
@@ -92,6 +92,8 @@ void au_cpup_igen(struct inode *inode, struct inode *h_inode)
 {
 	struct au_iinfo *iinfo = au_ii(inode);
 
+	IiMustWriteLock(inode);
+
 	iinfo->ii_higen = h_inode->i_generation;
 	iinfo->ii_hsb1 = h_inode->i_sb;
 }
@@ -139,7 +141,7 @@ void au_dtime_revert(struct au_dtime *dt)
 
 	err = vfsub_notify_change(&dt->dt_h_path, &attr);
 	if (unlikely(err))
-		AuWarn("restoring timestamps failed(%d). ignored\n", err);
+		pr_warning("restoring timestamps failed(%d). ignored\n", err);
 }
 
 /* ---------------------------------------------------------------------- */
@@ -150,21 +152,26 @@ int cpup_iattr(struct dentry *dst, aufs_bindex_t bindex, struct dentry *h_src)
 	int err, sbits;
 	struct iattr ia;
 	struct path h_path;
-	struct inode *h_isrc;
+	struct inode *h_isrc, *h_idst;
 
 	h_path.dentry = au_h_dptr(dst, bindex);
+	h_idst = h_path.dentry->d_inode;
 	h_path.mnt = au_sbr_mnt(dst->d_sb, bindex);
 	h_isrc = h_src->d_inode;
-	ia.ia_valid = ATTR_FORCE | ATTR_MODE | ATTR_UID | ATTR_GID
+	ia.ia_valid = ATTR_FORCE | ATTR_UID | ATTR_GID
 		| ATTR_ATIME | ATTR_MTIME
 		| ATTR_ATIME_SET | ATTR_MTIME_SET;
-	ia.ia_mode = h_isrc->i_mode;
 	ia.ia_uid = h_isrc->i_uid;
 	ia.ia_gid = h_isrc->i_gid;
 	ia.ia_atime = h_isrc->i_atime;
 	ia.ia_mtime = h_isrc->i_mtime;
-	sbits = !!(ia.ia_mode & (S_ISUID | S_ISGID));
-	au_cpup_attr_flags(h_path.dentry->d_inode, h_isrc);
+	if (h_idst->i_mode != h_isrc->i_mode
+	    && !S_ISLNK(h_idst->i_mode)) {
+		ia.ia_valid |= ATTR_MODE;
+		ia.ia_mode = h_isrc->i_mode;
+	}
+	sbits = !!(h_isrc->i_mode & (S_ISUID | S_ISGID));
+	au_cpup_attr_flags(h_idst, h_isrc);
 	err = vfsub_notify_change(&h_path, &ia);
 
 	/* is this nfs only? */
@@ -305,7 +312,7 @@ int au_copy_file(struct file *dst, struct file *src, loff_t len)
 	else
 		free_page((unsigned long)buf);
 
- out:
+out:
 	return err;
 }
 
@@ -314,7 +321,7 @@ int au_copy_file(struct file *dst, struct file *src, loff_t len)
  * we need to close the file.
  */
 static int au_cp_regular(struct dentry *dentry, aufs_bindex_t bdst,
-			aufs_bindex_t bsrc, loff_t len)
+			 aufs_bindex_t bsrc, loff_t len)
 {
 	int err, i;
 	enum { SRC, DST };
@@ -360,13 +367,13 @@ static int au_cp_regular(struct dentry *dentry, aufs_bindex_t bdst,
 	IMustLock(file[SRC].dentry->d_inode);
 	err = au_copy_file(file[DST].file, file[SRC].file, len);
 
- out_dst:
+out_dst:
 	fput(file[DST].file);
 	au_sbr_put(sb, file[DST].bindex);
- out_src:
+out_src:
 	fput(file[SRC].file);
 	au_sbr_put(sb, file[SRC].bindex);
- out:
+out:
 	return err;
 }
 
@@ -393,7 +400,7 @@ static int au_do_cpup_regular(struct dentry *dentry, aufs_bindex_t bdst,
 		err = -EIO;
 	}
 
- out:
+out:
 	return err;
 }
 
@@ -402,31 +409,33 @@ static int au_do_cpup_symlink(struct path *h_path, struct dentry *h_src,
 {
 	int err, symlen;
 	mm_segment_t old_fs;
-	char *sym;
+	union {
+		char *k;
+		char __user *u;
+	} sym;
 
 	err = -ENOSYS;
 	if (unlikely(!h_src->d_inode->i_op->readlink))
 		goto out;
 
 	err = -ENOMEM;
-	sym = __getname();
-	if (unlikely(!sym))
+	sym.k = __getname_gfp(GFP_NOFS);
+	if (unlikely(!sym.k))
 		goto out;
 
 	old_fs = get_fs();
 	set_fs(KERNEL_DS);
-	symlen = h_src->d_inode->i_op->readlink(h_src, (char __user *)sym,
-						PATH_MAX);
+	symlen = h_src->d_inode->i_op->readlink(h_src, sym.u, PATH_MAX);
 	err = symlen;
 	set_fs(old_fs);
 
 	if (symlen > 0) {
-		sym[symlen] = 0;
-		err = vfsub_symlink(h_dir, h_path, sym);
+		sym.k[symlen] = 0;
+		err = vfsub_symlink(h_dir, h_path, sym.k);
 	}
-	__putname(sym);
+	__putname(sym.k);
 
- out:
+out:
 	return err;
 }
 
@@ -518,7 +527,7 @@ int cpup_entry(struct dentry *dentry, aufs_bindex_t bdst,
 	    /* && dentry->d_inode->i_nlink == 1 */
 	    && bdst < bsrc
 	    && !au_ftest_cpup(flags, KEEPLINO))
-		au_xino_write0(sb, bsrc, h_inode->i_ino, /*ino*/0);
+		au_xino_write(sb, bsrc, h_inode->i_ino, /*ino*/0);
 		/* ignore this error */
 
 	if (do_dt)
@@ -567,8 +576,9 @@ static int au_cpup_single(struct dentry *dentry, aufs_bindex_t bdst,
 	if (dst_inode) {
 		if (unlikely(!plink)) {
 			err = -EIO;
-			AuIOErr("i%lu exists on a upper branch "
-				"but plink is disabled\n", inode->i_ino);
+			AuIOErr("hi%lu(i%lu) exists on b%d "
+				"but plink is disabled\n",
+				dst_inode->i_ino, inode->i_ino, bdst);
 			goto out;
 		}
 
@@ -582,7 +592,8 @@ static int au_cpup_single(struct dentry *dentry, aufs_bindex_t bdst,
 			if (unlikely(!h_src->d_inode)) {
 				err = -EIO;
 				AuIOErr("i%lu exists on a upper branch "
-					"but plink is broken\n", inode->i_ino);
+					"but not pseudo-linked\n",
+					inode->i_ino);
 				dput(h_src);
 				goto out;
 			}
@@ -600,7 +611,7 @@ static int au_cpup_single(struct dentry *dentry, aufs_bindex_t bdst,
 		} else
 			/* todo: cpup_wh_file? */
 			/* udba work */
-			au_update_brange(inode, 1);
+			au_update_ibrange(inode, /*do_put_zero*/1);
 	}
 
 	old_ibstart = au_ibstart(inode);
@@ -613,8 +624,14 @@ static int au_cpup_single(struct dentry *dentry, aufs_bindex_t bdst,
 	err = cpup_iattr(dentry, bdst, h_src);
 	isdir = S_ISDIR(dst_inode->i_mode);
 	if (!err) {
-		if (bdst < old_ibstart)
+		if (bdst < old_ibstart) {
+			if (S_ISREG(inode->i_mode)) {
+				err = au_dy_iaop(inode, bdst, dst_inode);
+				if (unlikely(err))
+					goto out_rev;
+			}
 			au_set_ibstart(inode, bdst);
+		}
 		au_set_h_iptr(inode, bdst, au_igrab(dst_inode),
 			      au_hi_flags(inode, isdir));
 		mutex_unlock(&dst_inode->i_mutex);
@@ -626,6 +643,7 @@ static int au_cpup_single(struct dentry *dentry, aufs_bindex_t bdst,
 	}
 
 	/* revert */
+out_rev:
 	h_path.dentry = h_parent;
 	mutex_unlock(&dst_inode->i_mutex);
 	au_dtime_store(&dt, dst_parent, &h_path);
@@ -640,7 +658,7 @@ static int au_cpup_single(struct dentry *dentry, aufs_bindex_t bdst,
 		err = -EIO;
 	}
 
- out:
+out:
 	dput(dst_parent);
 	return err;
 }
@@ -661,18 +679,46 @@ static void au_call_cpup_single(void *args)
 				  a->flags, a->dst_parent);
 }
 
+/*
+ * prevent SIGXFSZ in copy-up.
+ * testing CAP_MKNOD is for generic fs,
+ * but CAP_FSETID is for xfs only, currently.
+ */
+static int au_cpup_sio_test(struct super_block *sb, umode_t mode)
+{
+	int do_sio;
+
+	do_sio = 0;
+	if (!au_wkq_test()
+	    && (!au_sbi(sb)->si_plink_maint_pid
+		|| au_plink_maint(sb, AuLock_NOPLM))) {
+		switch (mode & S_IFMT) {
+		case S_IFREG:
+			/* no condition about RLIMIT_FSIZE and the file size */
+			do_sio = 1;
+			break;
+		case S_IFCHR:
+		case S_IFBLK:
+			do_sio = !capable(CAP_MKNOD);
+			break;
+		}
+		if (!do_sio)
+			do_sio = ((mode & (S_ISUID | S_ISGID))
+				  && !capable(CAP_FSETID));
+	}
+
+	return do_sio;
+}
+
 int au_sio_cpup_single(struct dentry *dentry, aufs_bindex_t bdst,
 		       aufs_bindex_t bsrc, loff_t len, unsigned int flags,
 		       struct dentry *dst_parent)
 {
 	int err, wkq_err;
-	umode_t mode;
 	struct dentry *h_dentry;
 
 	h_dentry = au_h_dptr(dentry, bsrc);
-	mode = h_dentry->d_inode->i_mode & S_IFMT;
-	if ((mode != S_IFCHR && mode != S_IFBLK)
-	    || capable(CAP_MKNOD))
+	if (!au_cpup_sio_test(dentry->d_sb, h_dentry->d_inode->i_mode))
 		err = au_cpup_single(dentry, bdst, bsrc, len, flags,
 				     dst_parent);
 	else {
@@ -740,25 +786,13 @@ int au_sio_cpup_simple(struct dentry *dentry, aufs_bindex_t bdst, loff_t len,
 		       unsigned int flags)
 {
 	int err, wkq_err;
-	unsigned char do_sio;
 	struct dentry *parent;
 	struct inode *h_dir;
 
 	parent = dget_parent(dentry);
 	h_dir = au_h_iptr(parent->d_inode, bdst);
-	do_sio = !!au_test_h_perm_sio(h_dir, MAY_EXEC | MAY_WRITE);
-	if (!do_sio) {
-		/*
-		 * testing CAP_MKNOD is for generic fs,
-		 * but CAP_FSETID is for xfs only, currently.
-		 */
-		umode_t mode = dentry->d_inode->i_mode;
-		do_sio = (((mode & (S_IFCHR | S_IFBLK))
-			   && !capable(CAP_MKNOD))
-			  || ((mode & (S_ISUID | S_ISGID))
-			      && !capable(CAP_FSETID)));
-	}
-	if (!do_sio)
+	if (!au_test_h_perm_sio(h_dir, MAY_EXEC | MAY_WRITE)
+	    && !au_cpup_sio_test(dentry->d_sb, dentry->d_inode->i_mode))
 		err = au_cpup_simple(dentry, bdst, len, flags);
 	else {
 		struct au_cpup_simple_args args = {
@@ -790,23 +824,28 @@ static int au_do_cpup_wh(struct dentry *dentry, aufs_bindex_t bdst,
 	aufs_bindex_t bstart;
 	struct au_dinfo *dinfo;
 	struct dentry *h_d_dst, *h_d_start;
+	struct au_hdentry *hdp;
 
 	dinfo = au_di(dentry);
+	AuRwMustWriteLock(&dinfo->di_rwsem);
+
 	bstart = dinfo->di_bstart;
-	h_d_dst = dinfo->di_hdentry[0 + bdst].hd_dentry;
+	hdp = dinfo->di_hdentry;
+	h_d_dst = hdp[0 + bdst].hd_dentry;
 	dinfo->di_bstart = bdst;
-	dinfo->di_hdentry[0 + bdst].hd_dentry = wh_dentry;
-	h_d_start = dinfo->di_hdentry[0 + bstart].hd_dentry;
-	if (file)
-		dinfo->di_hdentry[0 + bstart].hd_dentry
-			= au_h_fptr(file, au_fbstart(file))->f_dentry;
+	hdp[0 + bdst].hd_dentry = wh_dentry;
+	if (file) {
+		h_d_start = hdp[0 + bstart].hd_dentry;
+		hdp[0 + bstart].hd_dentry = au_hf_top(file)->f_dentry;
+	}
 	err = au_cpup_single(dentry, bdst, bstart, len, !AuCpup_DTIME,
 			     /*h_parent*/NULL);
-	if (!err && file) {
-		err = au_reopen_nondir(file);
-		dinfo->di_hdentry[0 + bstart].hd_dentry = h_d_start;
+	if (file) {
+		if (!err)
+			err = au_reopen_nondir(file);
+		hdp[0 + bstart].hd_dentry = h_d_start;
 	}
-	dinfo->di_hdentry[0 + bdst].hd_dentry = h_d_dst;
+	hdp[0 + bdst].hd_dentry = h_d_dst;
 	dinfo->di_bstart = bstart;
 
 	return err;
@@ -838,7 +877,10 @@ static int au_cpup_wh(struct dentry *dentry, aufs_bindex_t bdst, loff_t len,
 
 	dget(wh_dentry);
 	h_path.dentry = wh_dentry;
-	err = vfsub_unlink(h_parent->d_inode, &h_path, /*force*/0);
+	if (!S_ISDIR(wh_dentry->d_inode->i_mode))
+		err = vfsub_unlink(h_parent->d_inode, &h_path, /*force*/0);
+	else
+		err = vfsub_rmdir(h_parent->d_inode, &h_path);
 	if (unlikely(err)) {
 		AuIOErr("failed remove copied-up tmp file %.*s(%d)\n",
 			AuDLNPair(wh_dentry), err);
@@ -847,9 +889,9 @@ static int au_cpup_wh(struct dentry *dentry, aufs_bindex_t bdst, loff_t len,
 	au_dtime_revert(&dt);
 	au_set_hi_wh(dentry->d_inode, bdst, wh_dentry);
 
- out_wh:
+out_wh:
 	dput(wh_dentry);
- out:
+out:
 	dput(parent);
 	return err;
 }
@@ -887,25 +929,25 @@ int au_sio_cpup_wh(struct dentry *dentry, aufs_bindex_t bdst, loff_t len,
 		h_orph = wbr->wbr_orph;
 
 		h_parent = dget(au_h_dptr(parent, bdst));
-		au_set_h_dptr(parent, bdst, NULL);
 		au_set_h_dptr(parent, bdst, dget(h_orph));
 		h_tmpdir = h_orph->d_inode;
-		au_set_h_iptr(dir, bdst, NULL, 0);
 		au_set_h_iptr(dir, bdst, au_igrab(h_tmpdir), /*flags*/0);
 
 		/* this temporary unlock is safe */
 		if (file)
-			h_dentry = au_h_fptr(file, au_fbstart(file))->f_dentry;
+			h_dentry = au_hf_top(file)->f_dentry;
 		else
 			h_dentry = au_h_dptr(dentry, au_dbstart(dentry));
 		h_inode = h_dentry->d_inode;
 		IMustLock(h_inode);
 		mutex_unlock(&h_inode->i_mutex);
-		mutex_lock_nested(&h_tmpdir->i_mutex, AuLsc_I_PARENT2);
+		mutex_lock_nested(&h_tmpdir->i_mutex, AuLsc_I_PARENT3);
 		mutex_lock_nested(&h_inode->i_mutex, AuLsc_I_CHILD);
+		/* todo: au_h_open_pre()? */
 	}
 
-	if (!au_test_h_perm_sio(h_tmpdir, MAY_EXEC | MAY_WRITE))
+	if (!au_test_h_perm_sio(h_tmpdir, MAY_EXEC | MAY_WRITE)
+	    && !au_cpup_sio_test(dentry->d_sb, dentry->d_inode->i_mode))
 		err = au_cpup_wh(dentry, bdst, len, file);
 	else {
 		struct au_cpup_wh_args args = {
@@ -922,9 +964,8 @@ int au_sio_cpup_wh(struct dentry *dentry, aufs_bindex_t bdst, loff_t len,
 
 	if (h_orph) {
 		mutex_unlock(&h_tmpdir->i_mutex);
-		au_set_h_iptr(dir, bdst, NULL, 0);
+		/* todo: au_h_open_post()? */
 		au_set_h_iptr(dir, bdst, au_igrab(h_dir), /*flags*/0);
-		au_set_h_dptr(parent, bdst, NULL);
 		au_set_h_dptr(parent, bdst, h_parent);
 	}
 	iput(h_dir);
@@ -997,7 +1038,7 @@ int au_cp_dirs(struct dentry *dentry, aufs_bindex_t bdst,
 			break;
 	}
 
- out:
+out:
 	dput(parent);
 	return err;
 }
@@ -1033,7 +1074,7 @@ int au_test_and_cpup_dirs(struct dentry *dentry, aufs_bindex_t bdst)
 		err = au_cpup_dirs(dentry, bdst);
 	di_downgrade_lock(parent, AuLock_IR);
 
- out:
+out:
 	dput(parent);
 	return err;
 }
