@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2009 Junjiro R. Okajima
+ * Copyright (C) 2005-2011 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,9 +21,12 @@
  */
 
 #include <linux/buffer_head.h>
+#include <linux/jiffies.h>
 #include <linux/module.h>
 #include <linux/seq_file.h>
 #include <linux/statfs.h>
+#include <linux/vmalloc.h>
+#include <linux/writeback.h>
 #include "aufs.h"
 
 /*
@@ -35,7 +38,7 @@ static struct inode *aufs_alloc_inode(struct super_block *sb __maybe_unused)
 
 	c = au_cache_alloc_icntnr();
 	if (c) {
-		inode_init_once(&c->vfs_inode);
+		au_icntnr_init(c);
 		c->vfs_inode.i_version = 1; /* sigen(sb); */
 		c->iinfo.ii_hinode = NULL;
 		return &c->vfs_inode;
@@ -43,10 +46,18 @@ static struct inode *aufs_alloc_inode(struct super_block *sb __maybe_unused)
 	return NULL;
 }
 
+static void aufs_destroy_inode_cb(struct rcu_head *head)
+{
+	struct inode *inode = container_of(head, struct inode, i_rcu);
+
+	INIT_LIST_HEAD(&inode->i_dentry);
+	au_cache_free_icntnr(container_of(inode, struct au_icntnr, vfs_inode));
+}
+
 static void aufs_destroy_inode(struct inode *inode)
 {
 	au_iinfo_fin(inode);
-	au_cache_free_icntnr(container_of(inode, struct au_icntnr, vfs_inode));
+	call_rcu(&inode->i_rcu, aufs_destroy_inode_cb);
 }
 
 struct inode *au_iget_locked(struct super_block *sb, ino_t ino)
@@ -72,7 +83,7 @@ struct inode *au_iget_locked(struct super_block *sb, ino_t ino)
 		inode = ERR_PTR(err);
 	}
 
- out:
+out:
 	/* never return NULL */
 	AuDebugOn(!inode);
 	AuTraceErrPtr(inode);
@@ -85,16 +96,16 @@ static int au_show_brs(struct seq_file *seq, struct super_block *sb)
 	int err;
 	aufs_bindex_t bindex, bend;
 	struct path path;
-	struct au_hdentry *hd;
+	struct au_hdentry *hdp;
 	struct au_branch *br;
 
 	err = 0;
 	bend = au_sbend(sb);
-	hd = au_di(sb->s_root)->di_hdentry;
+	hdp = au_di(sb->s_root)->di_hdentry;
 	for (bindex = 0; !err && bindex <= bend; bindex++) {
 		br = au_sbr(sb, bindex);
 		path.mnt = br->br_mnt;
-		path.dentry = hd[bindex].hd_dentry;
+		path.dentry = hdp[bindex].hd_dentry;
 		err = au_seq_path(seq, &path);
 		if (err > 0)
 			err = seq_printf(seq, "=%s",
@@ -111,6 +122,8 @@ static void au_show_wbr_create(struct seq_file *m, int v,
 {
 	const char *pat;
 
+	AuRwMustAnyLock(&sbinfo->si_rwsem);
+
 	seq_printf(m, ",create=");
 	pat = au_optstr_wbr_create(v);
 	switch (v) {
@@ -122,11 +135,13 @@ static void au_show_wbr_create(struct seq_file *m, int v,
 		break;
 	case AuWbrCreate_MFSV:
 		seq_printf(m, /*pat*/"mfs:%lu",
-			   sbinfo->si_wbr_mfs.mfs_expire / HZ);
+			   jiffies_to_msecs(sbinfo->si_wbr_mfs.mfs_expire)
+			   / MSEC_PER_SEC);
 		break;
 	case AuWbrCreate_PMFSV:
 		seq_printf(m, /*pat*/"pmfs:%lu",
-			   sbinfo->si_wbr_mfs.mfs_expire / HZ);
+			   jiffies_to_msecs(sbinfo->si_wbr_mfs.mfs_expire)
+			   / MSEC_PER_SEC);
 		break;
 	case AuWbrCreate_MFSRR:
 		seq_printf(m, /*pat*/"mfsrr:%llu",
@@ -135,7 +150,8 @@ static void au_show_wbr_create(struct seq_file *m, int v,
 	case AuWbrCreate_MFSRRV:
 		seq_printf(m, /*pat*/"mfsrr:%llu:%lu",
 			   sbinfo->si_wbr_mfs.mfsrr_watermark,
-			   sbinfo->si_wbr_mfs.mfs_expire / HZ);
+			   jiffies_to_msecs(sbinfo->si_wbr_mfs.mfs_expire)
+			   / MSEC_PER_SEC);
 		break;
 	}
 }
@@ -152,6 +168,9 @@ static int au_show_xino(struct seq_file *seq, struct vfsmount *mnt)
 	struct qstr *name;
 	struct file *f;
 	struct dentry *d, *h_root;
+	struct au_hdentry *hdp;
+
+	AuRwMustAnyLock(&sbinfo->si_rwsem);
 
 	err = 0;
 	sb = mnt->mnt_sb;
@@ -164,7 +183,8 @@ static int au_show_xino(struct seq_file *seq, struct vfsmount *mnt)
 	brid = au_xino_brid(sb);
 	if (brid >= 0) {
 		bindex = au_br_index(sb, brid);
-		h_root = au_di(sb->s_root)->di_hdentry[0 + bindex].hd_dentry;
+		hdp = au_di(sb->s_root)->di_hdentry;
+		h_root = hdp[0 + bindex].hd_dentry;
 	}
 	d = f->f_dentry;
 	name = &d->d_name;
@@ -177,7 +197,7 @@ static int au_show_xino(struct seq_file *seq, struct vfsmount *mnt)
 	seq_puts(seq, ",xino=");
 	err = au_xino_path(seq, f);
 
- out:
+out:
 	return err;
 #endif
 }
@@ -185,7 +205,7 @@ static int au_show_xino(struct seq_file *seq, struct vfsmount *mnt)
 /* seq_file will re-call me in case of too long string */
 static int aufs_show_options(struct seq_file *m, struct vfsmount *mnt)
 {
-	int err, n;
+	int err;
 	unsigned int mnt_flags, v;
 	struct super_block *sb;
 	struct au_sbinfo *sbinfo;
@@ -225,6 +245,7 @@ static int aufs_show_options(struct seq_file *m, struct vfsmount *mnt)
 	AuStr(UDBA, udba);
 	AuBool(SHWH, shwh);
 	AuBool(PLINK, plink);
+	AuBool(DIO, dio);
 	/* AuBool(DIRPERM1, dirperm1); */
 	/* AuBool(REFROF, refrof); */
 
@@ -242,8 +263,8 @@ static int aufs_show_options(struct seq_file *m, struct vfsmount *mnt)
 
 	AuUInt(DIRWH, dirwh, sbinfo->si_dirwh);
 
-	n = sbinfo->si_rdcache / HZ;
-	AuUInt(RDCACHE, rdcache, n);
+	v = jiffies_to_msecs(sbinfo->si_rdcache) / MSEC_PER_SEC;
+	AuUInt(RDCACHE, rdcache, v);
 
 	AuUInt(RDBLK, rdblk, sbinfo->si_rdblk);
 	AuUInt(RDHASH, rdhash, sbinfo->si_rdhash);
@@ -253,7 +274,7 @@ static int aufs_show_options(struct seq_file *m, struct vfsmount *mnt)
 	AuBool(WARN_PERM, warn_perm);
 	AuBool(VERBOSE, verbose);
 
- out:
+out:
 	/* be sure to print "br:" last */
 	if (!sysaufs_brs) {
 		seq_puts(m, ",br:");
@@ -262,9 +283,9 @@ static int aufs_show_options(struct seq_file *m, struct vfsmount *mnt)
 	si_read_unlock(sb);
 	return 0;
 
-#undef Deleted
 #undef AuBool
 #undef AuStr
+#undef AuUInt
 }
 
 /* ---------------------------------------------------------------------- */
@@ -288,7 +309,7 @@ static int au_statfs_sum(struct super_block *sb, struct kstatfs *buf)
 	u64 blocks, bfree, bavail, files, ffree;
 	aufs_bindex_t bend, bindex, i;
 	unsigned char shared;
-	struct vfsmount *h_mnt;
+	struct path h_path;
 	struct super_block *h_sb;
 
 	blocks = 0;
@@ -300,8 +321,8 @@ static int au_statfs_sum(struct super_block *sb, struct kstatfs *buf)
 	err = 0;
 	bend = au_sbend(sb);
 	for (bindex = bend; bindex >= 0; bindex--) {
-		h_mnt = au_sbr_mnt(sb, bindex);
-		h_sb = h_mnt->mnt_sb;
+		h_path.mnt = au_sbr_mnt(sb, bindex);
+		h_sb = h_path.mnt->mnt_sb;
 		shared = 0;
 		for (i = bindex + 1; !shared && i <= bend; i++)
 			shared = (au_sbr_sb(sb, i) == h_sb);
@@ -309,7 +330,8 @@ static int au_statfs_sum(struct super_block *sb, struct kstatfs *buf)
 			continue;
 
 		/* sb->s_root for NFS is unreliable */
-		err = vfs_statfs(h_mnt->mnt_root, buf);
+		h_path.dentry = h_path.mnt->mnt_root;
+		err = vfs_statfs(&h_path, buf);
 		if (unlikely(err))
 			goto out;
 
@@ -326,28 +348,31 @@ static int au_statfs_sum(struct super_block *sb, struct kstatfs *buf)
 	buf->f_files = files;
 	buf->f_ffree = ffree;
 
- out:
+out:
 	return err;
 }
 
 static int aufs_statfs(struct dentry *dentry, struct kstatfs *buf)
 {
 	int err;
+	struct path h_path;
 	struct super_block *sb;
 
 	/* lock free root dinfo */
 	sb = dentry->d_sb;
 	si_noflush_read_lock(sb);
-	if (!au_opt_test(au_mntflags(sb), SUM))
+	if (!au_opt_test(au_mntflags(sb), SUM)) {
 		/* sb->s_root for NFS is unreliable */
-		err = vfs_statfs(au_sbr_mnt(sb, 0)->mnt_root, buf);
-	else
+		h_path.mnt = au_sbr_mnt(sb, 0);
+		h_path.dentry = h_path.mnt->mnt_root;
+		err = vfs_statfs(&h_path, buf);
+	} else
 		err = au_statfs_sum(sb, buf);
 	si_read_unlock(sb);
 
 	if (!err) {
 		buf->f_type = AUFS_SUPER_MAGIC;
-		buf->f_namelen -= AUFS_WH_PFX_LEN;
+		buf->f_namelen = AUFS_MAX_NAMELEN;
 		memset(&buf->f_fsid, 0, sizeof(buf->f_fsid));
 	}
 	/* buf->f_bsize = buf->f_blocks = buf->f_bfree = buf->f_bavail = -1; */
@@ -357,55 +382,6 @@ static int aufs_statfs(struct dentry *dentry, struct kstatfs *buf)
 
 /* ---------------------------------------------------------------------- */
 
-/* try flushing the lower fs at aufs remount/unmount time */
-
-static void au_fsync_br(struct super_block *sb)
-{
-	aufs_bindex_t bend, bindex;
-	int brperm;
-	struct au_branch *br;
-	struct super_block *h_sb;
-
-	bend = au_sbend(sb);
-	for (bindex = 0; bindex < bend; bindex++) {
-		br = au_sbr(sb, bindex);
-		brperm = br->br_perm;
-		if (brperm == AuBrPerm_RR || brperm == AuBrPerm_RRWH)
-			continue;
-		h_sb = br->br_mnt->mnt_sb;
-		if (bdev_read_only(h_sb->s_bdev))
-			continue;
-
-		lockdep_off();
-		down_write(&h_sb->s_umount);
-		shrink_dcache_sb(h_sb);
-		fsync_super(h_sb);
-		up_write(&h_sb->s_umount);
-		lockdep_on();
-	}
-}
-
-/*
- * this IS NOT for super_operations.
- * I guess it will be reverted someday.
- */
-static void aufs_umount_begin(struct super_block *sb)
-{
-	struct au_sbinfo *sbinfo;
-
-	sbinfo = au_sbi(sb);
-	if (!sbinfo)
-		return;
-
-	si_write_lock(sb);
-	au_fsync_br(sb);
-	if (au_opt_test(au_mntflags(sb), PLINK))
-		au_plink_put(sb);
-	if (sbinfo->si_wbr_create_ops->fin)
-		sbinfo->si_wbr_create_ops->fin(sb);
-	si_write_unlock(sb);
-}
-
 /* final actions when unmounting a file system */
 static void aufs_put_super(struct super_block *sb)
 {
@@ -415,216 +391,267 @@ static void aufs_put_super(struct super_block *sb)
 	if (!sbinfo)
 		return;
 
-	aufs_umount_begin(sb);
 	dbgaufs_si_fin(sbinfo);
 	kobject_put(&sbinfo->si_kobj);
 }
 
 /* ---------------------------------------------------------------------- */
 
+void au_array_free(void *array)
+{
+	if (array) {
+		if (!is_vmalloc_addr(array))
+			kfree(array);
+		else
+			vfree(array);
+	}
+}
+
+void *au_array_alloc(unsigned long long *hint, au_arraycb_t cb, void *arg)
+{
+	void *array;
+	unsigned long long n;
+
+	array = NULL;
+	n = 0;
+	if (!*hint)
+		goto out;
+
+	if (*hint > ULLONG_MAX / sizeof(array)) {
+		array = ERR_PTR(-EMFILE);
+		pr_err("hint %llu\n", *hint);
+		goto out;
+	}
+
+	array = kmalloc(sizeof(array) * *hint, GFP_NOFS);
+	if (unlikely(!array))
+		array = vmalloc(sizeof(array) * *hint);
+	if (unlikely(!array)) {
+		array = ERR_PTR(-ENOMEM);
+		goto out;
+	}
+
+	n = cb(array, *hint, arg);
+	AuDebugOn(n > *hint);
+
+out:
+	*hint = n;
+	return array;
+}
+
+static unsigned long long au_iarray_cb(void *a,
+				       unsigned long long max __maybe_unused,
+				       void *arg)
+{
+	unsigned long long n;
+	struct inode **p, *inode;
+	struct list_head *head;
+
+	n = 0;
+	p = a;
+	head = arg;
+	spin_lock(&inode_sb_list_lock);
+	list_for_each_entry(inode, head, i_sb_list) {
+		if (!is_bad_inode(inode)
+		    && au_ii(inode)->ii_bstart >= 0) {
+			spin_lock(&inode->i_lock);
+			if (atomic_read(&inode->i_count)) {
+				au_igrab(inode);
+				*p++ = inode;
+				n++;
+				AuDebugOn(n > max);
+			}
+			spin_unlock(&inode->i_lock);
+		}
+	}
+	spin_unlock(&inode_sb_list_lock);
+
+	return n;
+}
+
+struct inode **au_iarray_alloc(struct super_block *sb, unsigned long long *max)
+{
+	*max = atomic_long_read(&au_sbi(sb)->si_ninodes);
+	return au_array_alloc(max, au_iarray_cb, &sb->s_inodes);
+}
+
+void au_iarray_free(struct inode **a, unsigned long long max)
+{
+	unsigned long long ull;
+
+	for (ull = 0; ull < max; ull++)
+		iput(a[ull]);
+	au_array_free(a);
+}
+
+/* ---------------------------------------------------------------------- */
+
 /*
  * refresh dentry and inode at remount time.
  */
-static int do_refresh(struct dentry *dentry, mode_t type,
-		      unsigned int dir_flags)
+/* todo: consolidate with simple_reval_dpath() and au_reval_for_attr() */
+static int au_do_refresh(struct dentry *dentry, unsigned int dir_flags,
+		      struct dentry *parent)
 {
 	int err;
-	struct dentry *parent;
 
 	di_write_lock_child(dentry);
-	parent = dget_parent(dentry);
 	di_read_lock_parent(parent, AuLock_IR);
-
-	/* returns the number of positive dentries */
-	err = au_refresh_hdentry(dentry, type);
-	if (err >= 0) {
-		struct inode *inode = dentry->d_inode;
-		err = au_refresh_hinode(inode, dentry);
-		if (!err && type == S_IFDIR)
-			au_reset_hinotify(inode, dir_flags);
-	}
-	if (unlikely(err))
-		AuErr("unrecoverable error %d, %.*s\n", err, AuDLNPair(dentry));
-
+	err = au_refresh_dentry(dentry, parent);
+	if (!err && dir_flags)
+		au_hn_reset(dentry->d_inode, dir_flags);
 	di_read_unlock(parent, AuLock_IR);
-	dput(parent);
 	di_write_unlock(dentry);
 
 	return err;
 }
 
-static int test_dir(struct dentry *dentry, void *arg __maybe_unused)
+static int au_do_refresh_d(struct dentry *dentry, unsigned int sigen,
+			   struct au_sbinfo *sbinfo,
+			   const unsigned int dir_flags)
 {
-	return S_ISDIR(dentry->d_inode->i_mode);
-}
-
-/* gave up consolidating with refresh_nondir() */
-static int refresh_dir(struct dentry *root, unsigned int sigen)
-{
-	int err, i, j, ndentry, e;
-	struct au_dcsub_pages dpages;
-	struct au_dpage *dpage;
-	struct dentry **dentries;
+	int err;
+	struct dentry *parent;
 	struct inode *inode;
-	const unsigned int flags = au_hi_flags(root->d_inode, /*isdir*/1);
 
 	err = 0;
-	list_for_each_entry(inode, &root->d_sb->s_inodes, i_sb_list)
-		if (S_ISDIR(inode->i_mode) && au_iigen(inode) != sigen) {
-			ii_write_lock_child(inode);
-			e = au_refresh_hinode_self(inode, /*do_attr*/1);
-			ii_write_unlock(inode);
-			if (unlikely(e)) {
-				AuDbg("e %d, i%lu\n", e, inode->i_ino);
-				if (!err)
-					err = e;
-				/* go on even if err */
+	parent = dget_parent(dentry);
+	if (!au_digen_test(parent, sigen) && au_digen_test(dentry, sigen)) {
+		inode = dentry->d_inode;
+		if (inode) {
+			if (!S_ISDIR(inode->i_mode))
+				err = au_do_refresh(dentry, /*dir_flags*/0,
+						 parent);
+			else {
+				err = au_do_refresh(dentry, dir_flags, parent);
+				if (unlikely(err))
+					au_fset_si(sbinfo, FAILED_REFRESH_DIR);
 			}
-		}
-
-	e = au_dpages_init(&dpages, GFP_NOFS);
-	if (unlikely(e)) {
-		if (!err)
-			err = e;
-		goto out;
-	}
-	e = au_dcsub_pages(&dpages, root, test_dir, NULL);
-	if (unlikely(e)) {
-		if (!err)
-			err = e;
-		goto out_dpages;
+		} else
+			err = au_do_refresh(dentry, /*dir_flags*/0, parent);
+		AuDbgDentry(dentry);
 	}
+	dput(parent);
 
-	for (i = 0; !e && i < dpages.ndpage; i++) {
-		dpage = dpages.dpages + i;
-		dentries = dpage->dentries;
-		ndentry = dpage->ndentry;
-		for (j = 0; !e && j < ndentry; j++) {
-			struct dentry *d;
-
-			d = dentries[j];
-			au_dbg_verify_dir_parent(d, sigen);
-			if (au_digen(d) != sigen) {
-				e = do_refresh(d, S_IFDIR, flags);
-				if (unlikely(e && !err))
-					err = e;
-				/* break on err */
-			}
-		}
-	}
-
- out_dpages:
-	au_dpages_free(&dpages);
- out:
+	AuTraceErr(err);
 	return err;
 }
 
-static int test_nondir(struct dentry *dentry, void *arg __maybe_unused)
-{
-	return !S_ISDIR(dentry->d_inode->i_mode);
-}
-
-static int refresh_nondir(struct dentry *root, unsigned int sigen,
-			  int do_dentry)
+static int au_refresh_d(struct super_block *sb)
 {
 	int err, i, j, ndentry, e;
+	unsigned int sigen;
 	struct au_dcsub_pages dpages;
 	struct au_dpage *dpage;
-	struct dentry **dentries;
-	struct inode *inode;
+	struct dentry **dentries, *d;
+	struct au_sbinfo *sbinfo;
+	struct dentry *root = sb->s_root;
+	const unsigned int dir_flags = au_hi_flags(root->d_inode, /*isdir*/1);
 
-	err = 0;
-	list_for_each_entry(inode, &root->d_sb->s_inodes, i_sb_list)
-		if (!S_ISDIR(inode->i_mode) && au_iigen(inode) != sigen) {
-			ii_write_lock_child(inode);
-			e = au_refresh_hinode_self(inode, /*do_attr*/1);
-			ii_write_unlock(inode);
-			if (unlikely(e)) {
-				AuDbg("e %d, i%lu\n", e, inode->i_ino);
-				if (!err)
-					err = e;
-				/* go on even if err */
-			}
-		}
-
-	if (!do_dentry)
+	err = au_dpages_init(&dpages, GFP_NOFS);
+	if (unlikely(err))
 		goto out;
-
-	e = au_dpages_init(&dpages, GFP_NOFS);
-	if (unlikely(e)) {
-		if (!err)
-			err = e;
-		goto out;
-	}
-	e = au_dcsub_pages(&dpages, root, test_nondir, NULL);
-	if (unlikely(e)) {
-		if (!err)
-			err = e;
+	err = au_dcsub_pages(&dpages, root, NULL, NULL);
+	if (unlikely(err))
 		goto out_dpages;
-	}
 
+	sigen = au_sigen(sb);
+	sbinfo = au_sbi(sb);
 	for (i = 0; i < dpages.ndpage; i++) {
 		dpage = dpages.dpages + i;
 		dentries = dpage->dentries;
 		ndentry = dpage->ndentry;
 		for (j = 0; j < ndentry; j++) {
-			struct dentry *d;
-
 			d = dentries[j];
-			au_dbg_verify_nondir_parent(d, sigen);
-			inode = d->d_inode;
-			if (inode && au_digen(d) != sigen) {
-				e = do_refresh(d, inode->i_mode & S_IFMT,
-					       /*dir_flags*/0);
-				if (unlikely(e && !err))
+			e = au_do_refresh_d(d, sigen, sbinfo, dir_flags);
+			if (unlikely(e && !err))
+				err = e;
+			/* go on even err */
+		}
+	}
+
+out_dpages:
+	au_dpages_free(&dpages);
+out:
+	return err;
+}
+
+static int au_refresh_i(struct super_block *sb)
+{
+	int err, e;
+	unsigned int sigen;
+	unsigned long long max, ull;
+	struct inode *inode, **array;
+
+	array = au_iarray_alloc(sb, &max);
+	err = PTR_ERR(array);
+	if (IS_ERR(array))
+		goto out;
+
+	err = 0;
+	sigen = au_sigen(sb);
+	for (ull = 0; ull < max; ull++) {
+		inode = array[ull];
+		if (au_iigen(inode) != sigen) {
+			ii_write_lock_child(inode);
+			e = au_refresh_hinode_self(inode);
+			ii_write_unlock(inode);
+			if (unlikely(e)) {
+				pr_err("error %d, i%lu\n", e, inode->i_ino);
+				if (!err)
 					err = e;
-				/* go on even err */
+				/* go on even if err */
 			}
 		}
 	}
 
- out_dpages:
-	au_dpages_free(&dpages);
- out:
+	au_iarray_free(array, max);
+
+out:
 	return err;
 }
 
-static void au_remount_refresh(struct super_block *sb, unsigned int flags)
+static void au_remount_refresh(struct super_block *sb)
 {
-	int err;
-	unsigned int sigen;
-	struct au_sbinfo *sbinfo;
+	int err, e;
+	unsigned int udba;
+	aufs_bindex_t bindex, bend;
 	struct dentry *root;
 	struct inode *inode;
+	struct au_branch *br;
 
 	au_sigen_inc(sb);
-	sigen = au_sigen(sb);
-	sbinfo = au_sbi(sb);
-	au_fclr_si(sbinfo, FAILED_REFRESH_DIRS);
+	au_fclr_si(au_sbi(sb), FAILED_REFRESH_DIR);
 
 	root = sb->s_root;
 	DiMustNoWaiters(root);
 	inode = root->d_inode;
 	IiMustNoWaiters(inode);
-	au_reset_hinotify(inode, au_hi_flags(inode, /*isdir*/1));
-	di_write_unlock(root);
 
-	err = refresh_dir(root, sigen);
-	if (unlikely(err)) {
-		au_fset_si(sbinfo, FAILED_REFRESH_DIRS);
-		AuWarn("Refreshing directories failed, ignored (%d)\n", err);
-	}
-
-	if (au_ftest_opts(flags, REFRESH_NONDIR)) {
-		err = refresh_nondir(root, sigen, !err);
+	udba = au_opt_udba(sb);
+	bend = au_sbend(sb);
+	for (bindex = 0; bindex <= bend; bindex++) {
+		br = au_sbr(sb, bindex);
+		err = au_hnotify_reset_br(udba, br, br->br_perm);
 		if (unlikely(err))
-			AuWarn("Refreshing non-directories failed, ignored"
-			       "(%d)\n", err);
+			AuIOErr("hnotify failed on br %d, %d, ignored\n",
+				bindex, err);
+		/* go on even if err */
 	}
+	au_hn_reset(inode, au_hi_flags(inode, /*isdir*/1));
 
+	di_write_unlock(root);
+	err = au_refresh_d(sb);
+	e = au_refresh_i(sb);
+	if (unlikely(e && !err))
+		err = e;
 	/* aufs_write_lock() calls ..._child() */
 	di_write_lock_child(root);
-	au_cpup_attr_all(root->d_inode, /*force*/1);
+
+	au_cpup_attr_all(inode, /*force*/1);
+
+	if (unlikely(err))
+		AuIOErr("refresh failed, ignored, %d\n", err);
 }
 
 /* stop extra interpretation of errno in mount(8), and strange error messages */
@@ -644,7 +671,8 @@ static int cvt_err(int err)
 
 static int aufs_remount_fs(struct super_block *sb, int *flags, char *data)
 {
-	int err;
+	int err, do_dx;
+	unsigned int mntflags;
 	struct au_opts opts;
 	struct dentry *root;
 	struct inode *inode;
@@ -653,11 +681,12 @@ static int aufs_remount_fs(struct super_block *sb, int *flags, char *data)
 	err = 0;
 	root = sb->s_root;
 	if (!data || !*data) {
-		aufs_write_lock(root);
-		err = au_opts_verify(sb, *flags, /*pending*/0);
-		if (!err)
-			au_fsync_br(sb);
-		aufs_write_unlock(root);
+		err = si_write_lock(sb, AuLock_FLUSH | AuLock_NOPLM);
+		if (!err) {
+			di_write_lock_child(root);
+			err = au_opts_verify(sb, *flags, /*pending*/0);
+			aufs_write_unlock(root);
+		}
 		goto out;
 	}
 
@@ -678,31 +707,40 @@ static int aufs_remount_fs(struct super_block *sb, int *flags, char *data)
 	sbinfo = au_sbi(sb);
 	inode = root->d_inode;
 	mutex_lock(&inode->i_mutex);
-	aufs_write_lock(root);
-	au_fsync_br(sb);
+	err = si_write_lock(sb, AuLock_FLUSH | AuLock_NOPLM);
+	if (unlikely(err))
+		goto out_mtx;
+	di_write_lock_child(root);
 
 	/* au_opts_remount() may return an error */
 	err = au_opts_remount(sb, &opts);
 	au_opts_free(&opts);
 
-	if (au_ftest_opts(opts.flags, REFRESH_DIR)
-	    || au_ftest_opts(opts.flags, REFRESH_NONDIR))
-		au_remount_refresh(sb, opts.flags);
+	if (au_ftest_opts(opts.flags, REFRESH))
+		au_remount_refresh(sb);
+
+	if (au_ftest_opts(opts.flags, REFRESH_DYAOP)) {
+		mntflags = au_mntflags(sb);
+		do_dx = !!au_opt_test(mntflags, DIO);
+		au_dy_arefresh(do_dx);
+	}
 
 	aufs_write_unlock(root);
-	mutex_unlock(&inode->i_mutex);
 
- out_opts:
+out_mtx:
+	mutex_unlock(&inode->i_mutex);
+out_opts:
 	free_page((unsigned long)opts.opt);
- out:
+out:
 	err = cvt_err(err);
 	AuTraceErr(err);
 	return err;
 }
 
-static struct super_operations aufs_sop = {
+static const struct super_operations aufs_sop = {
 	.alloc_inode	= aufs_alloc_inode,
 	.destroy_inode	= aufs_destroy_inode,
+	/* always deleting, no clearing */
 	.drop_inode	= generic_delete_inode,
 	.show_options	= aufs_show_options,
 	.statfs		= aufs_statfs,
@@ -737,7 +775,7 @@ static int alloc_root(struct super_block *sb)
 	if (IS_ERR(root))
 		goto out_iput;
 
-	err = au_alloc_dinfo(root);
+	err = au_di_init(root);
 	if (!err) {
 		sb->s_root = root;
 		return 0; /* success */
@@ -745,10 +783,9 @@ static int alloc_root(struct super_block *sb)
 	dput(root);
 	goto out; /* do not iput */
 
- out_iput:
+out_iput:
 	iget_failed(inode);
-	iput(inode);
- out:
+out:
 	return err;
 
 }
@@ -764,7 +801,7 @@ static int aufs_fill_super(struct super_block *sb, void *raw_data,
 
 	if (unlikely(!arg || !*arg)) {
 		err = -EINVAL;
-		AuErr("no arg\n");
+		pr_err("no arg\n");
 		goto out;
 	}
 
@@ -783,6 +820,7 @@ static int aufs_fill_super(struct super_block *sb, void *raw_data,
 	/* all timestamps always follow the ones on the branch */
 	sb->s_flags |= MS_NOATIME | MS_NODIRATIME;
 	sb->s_op = &aufs_sop;
+	sb->s_d_op = &aufs_dop;
 	sb->s_magic = AUFS_SUPER_MAGIC;
 	sb->s_maxbytes = 0;
 	au_export_init(sb);
@@ -808,29 +846,24 @@ static int aufs_fill_super(struct super_block *sb, void *raw_data,
 
 	/* lock vfs_inode first, then aufs. */
 	mutex_lock(&inode->i_mutex);
-	inode->i_op = &aufs_dir_iop;
-	inode->i_fop = &aufs_dir_fop;
 	aufs_write_lock(root);
 	err = au_opts_mount(sb, &opts);
 	au_opts_free(&opts);
-	if (unlikely(err))
-		goto out_unlock;
 	aufs_write_unlock(root);
 	mutex_unlock(&inode->i_mutex);
-	goto out_opts; /* success */
+	if (!err)
+		goto out_opts; /* success */
 
- out_unlock:
-	aufs_write_unlock(root);
-	mutex_unlock(&inode->i_mutex);
- out_root:
+out_root:
 	dput(root);
 	sb->s_root = NULL;
- out_info:
+out_info:
+	dbgaufs_si_fin(au_sbi(sb));
 	kobject_put(&au_sbi(sb)->si_kobj);
 	sb->s_fs_info = NULL;
- out_opts:
+out_opts:
 	free_page((unsigned long)opts.opt);
- out:
+out:
 	AuTraceErr(err);
 	err = cvt_err(err);
 	AuTraceErr(err);
@@ -839,23 +872,50 @@ static int aufs_fill_super(struct super_block *sb, void *raw_data,
 
 /* ---------------------------------------------------------------------- */
 
-static int aufs_get_sb(struct file_system_type *fs_type, int flags,
-		       const char *dev_name __maybe_unused, void *raw_data,
-		       struct vfsmount *mnt)
+static struct dentry *aufs_mount(struct file_system_type *fs_type, int flags,
+				 const char *dev_name __maybe_unused,
+				 void *raw_data)
 {
-	int err;
+	struct dentry *root;
 	struct super_block *sb;
 
 	/* all timestamps always follow the ones on the branch */
 	/* mnt->mnt_flags |= MNT_NOATIME | MNT_NODIRATIME; */
-	err = get_sb_nodev(fs_type, flags, raw_data, aufs_fill_super, mnt);
-	if (!err) {
-		sb = mnt->mnt_sb;
-		si_write_lock(sb);
-		sysaufs_brs_add(sb, 0);
-		si_write_unlock(sb);
+	root = mount_nodev(fs_type, flags, raw_data, aufs_fill_super);
+	if (IS_ERR(root))
+		goto out;
+
+	sb = root->d_sb;
+	si_write_lock(sb, !AuLock_FLUSH);
+	sysaufs_brs_add(sb, 0);
+	si_write_unlock(sb);
+	au_sbilist_add(sb);
+
+out:
+	return root;
+}
+
+static void aufs_kill_sb(struct super_block *sb)
+{
+	struct au_sbinfo *sbinfo;
+
+	sbinfo = au_sbi(sb);
+	if (sbinfo) {
+		au_sbilist_del(sb);
+		aufs_write_lock(sb->s_root);
+		if (sbinfo->si_wbr_create_ops->fin)
+			sbinfo->si_wbr_create_ops->fin(sb);
+		if (au_opt_test(sbinfo->si_mntflags, UDBA_HNOTIFY)) {
+			au_opt_set_udba(sbinfo->si_mntflags, UDBA_NONE);
+			au_remount_refresh(sb);
+		}
+		if (au_opt_test(sbinfo->si_mntflags, PLINK))
+			au_plink_put(sb, /*verbose*/1);
+		au_xino_clr(sb);
+		aufs_write_unlock(sb->s_root);
+		au_nwt_flush(&sbinfo->si_nowait);
 	}
-	return err;
+	generic_shutdown_super(sb);
 }
 
 struct file_system_type aufs_fs_type = {
@@ -863,8 +923,8 @@ struct file_system_type aufs_fs_type = {
 	.fs_flags	=
 		FS_RENAME_DOES_D_MOVE	/* a race between rename and others */
 		| FS_REVAL_DOT,		/* for NFS branch and udba */
-	.get_sb		= aufs_get_sb,
-	.kill_sb	= generic_shutdown_super,
+	.mount		= aufs_mount,
+	.kill_sb	= aufs_kill_sb,
 	/* no need to __module_get() and module_put(). */
 	.owner		= THIS_MODULE,
 };
