@@ -1,7 +1,7 @@
 /*
  * kernel/power/tuxonice.h
  *
- * Copyright (C) 2004-2007 Nigel Cunningham (nigel at tuxonice net)
+ * Copyright (C) 2004-2010 Nigel Cunningham (nigel at tuxonice net)
  *
  * This file is released under the GPLv2.
  *
@@ -15,24 +15,28 @@
 #include <linux/delay.h>
 #include <linux/bootmem.h>
 #include <linux/suspend.h>
-#include <linux/dyn_pageflags.h>
 #include <linux/fs.h>
 #include <linux/kmod.h>
+#include <linux/tuxonice.h>
 #include <asm/setup.h>
 #include "tuxonice_pageflags.h"
+#include "power.h"
 
-#define TOI_CORE_VERSION "3.0-rc5"
-
-#define MY_BOOT_KERNEL_DATA_VERSION 1
+#define TOI_CORE_VERSION "3.2"
+#define	TOI_HEADER_VERSION 3
+#define MY_BOOT_KERNEL_DATA_VERSION 3
 
 struct toi_boot_kernel_data {
 	int version;
 	int size;
 	unsigned long toi_action;
 	unsigned long toi_debug_state;
-	int toi_default_console_level;
+	u32 toi_default_console_level;
 	int toi_io_time[2][2];
 	char toi_nosave_commandline[COMMAND_LINE_SIZE];
+	unsigned long pages_used[33];
+	unsigned long compress_bytes_in;
+	unsigned long compress_bytes_out;
 };
 
 extern struct toi_boot_kernel_data toi_bkd;
@@ -45,7 +49,6 @@ extern unsigned long boot_kernel_data_buffer;
 enum {
 	TOI_REBOOT,
 	TOI_PAUSE,
-	TOI_SLOW,
 	TOI_LOGALL,
 	TOI_CAN_CANCEL,
 	TOI_KEEP_IMAGE,
@@ -55,19 +58,23 @@ enum {
 	TOI_TEST_FILTER_SPEED,
 	TOI_TEST_BIO,
 	TOI_NO_PAGESET2,
-	TOI_PM_PREPARE_CONSOLE,
 	TOI_IGNORE_ROOTFS,
 	TOI_REPLACE_SWSUSP,
 	TOI_PAGESET2_FULL,
 	TOI_ABORT_ON_RESAVE_NEEDED,
 	TOI_NO_MULTITHREADED_IO,
-	TOI_NO_DIRECT_LOAD,
+	TOI_NO_DIRECT_LOAD, /* Obsolete */
 	TOI_LATE_CPU_HOTPLUG,
-	TOI_GET_MAX_MEM_ALLOCD
+	TOI_GET_MAX_MEM_ALLOCD,
+	TOI_NO_FLUSHER_THREAD,
+	TOI_NO_PS2_IF_UNNEEDED,
+	TOI_POST_RESUME_BREAKPOINT,
+	TOI_NO_READAHEAD,
 };
 
+extern unsigned long toi_bootflags_mask;
+
 #define clear_action_state(bit) (test_and_clear_bit(bit, &toi_bkd.toi_action))
-#define test_action_state(bit) (test_bit(bit, &toi_bkd.toi_action))
 
 /*		 == Result states == 		*/
 
@@ -82,6 +89,7 @@ enum {
 	TOI_UNABLE_TO_FREE_ENOUGH_MEMORY,
 	TOI_PM_SEM,
 	TOI_DEVICE_REFUSED,
+	TOI_SYSDEV_REFUSED,
 	TOI_EXTRA_PAGES_ALLOW_TOO_SMALL,
 	TOI_UNABLE_TO_PREPARE_IMAGE,
 	TOI_FAILED_MODULE_INIT,
@@ -95,6 +103,14 @@ enum {
 	TOI_RESAVE_NEEDED,
 	TOI_CANT_SUSPEND,
 	TOI_NOTIFIERS_PREPARE_FAILED,
+	TOI_PRE_SNAPSHOT_FAILED,
+	TOI_PRE_RESTORE_FAILED,
+	TOI_USERMODE_HELPERS_ERR,
+	TOI_CANT_USE_ALT_RESUME,
+	TOI_HEADER_TOO_BIG,
+	TOI_WAKEUP_EVENT,
+	TOI_SYSCORE_REFUSED,
+	TOI_NUM_RESULT_STATES	/* Used in printing debug info only */
 };
 
 extern unsigned long toi_result;
@@ -124,10 +140,14 @@ enum {
 	TOI_HEADER,
 	TOI_WRITER,
 	TOI_MEMORY,
+	TOI_PAGEDIR,
+	TOI_COMPRESS,
+	TOI_BIO,
 };
 
 #define set_debug_state(bit) (test_and_set_bit(bit, &toi_bkd.toi_debug_state))
-#define clear_debug_state(bit) (test_and_clear_bit(bit, &toi_bkd.toi_debug_state))
+#define clear_debug_state(bit) \
+	(test_and_clear_bit(bit, &toi_bkd.toi_debug_state))
 #define test_debug_state(bit) (test_bit(bit, &toi_bkd.toi_debug_state))
 
 /*		== Steps in hibernating ==	*/
@@ -158,7 +178,7 @@ enum {
 struct toi_core_fns {
 	int (*post_context_save)(void);
 	unsigned long (*get_nonconflicting_page)(void);
-	int (*try_hibernate)(int have_pmsem);
+	int (*try_hibernate)(void);
 	void (*try_resume)(void);
 };
 
@@ -174,8 +194,8 @@ extern void toi_finish_anything(int toi_or_resume);
 extern int save_image_part1(void);
 extern int toi_atomic_restore(void);
 
-extern int _toi_try_hibernate(int have_pmsem);
-extern void __toi_try_resume(void);
+extern int toi_try_hibernate(void);
+extern void toi_try_resume(void);
 
 extern int __toi_post_context_save(void);
 
@@ -184,22 +204,21 @@ extern char alt_resume_param[256];
 
 extern void copyback_post(void);
 extern int toi_hibernate(void);
-extern int extra_pd1_pages_used;
+extern unsigned long extra_pd1_pages_used;
 
 #define SECTOR_SIZE 512
 
 extern void toi_early_boot_message(int can_erase_image, int default_answer,
 	char *warning_reason, ...);
 
-static inline int load_direct(struct page *page)
-{
-	return test_action_state(TOI_NO_DIRECT_LOAD) ? 0 :
-		PagePageset1Copy(page);
-}
-
-extern int pre_resume_freeze(void);
 extern int do_check_can_resume(void);
 extern int do_toi_step(int step);
 extern int toi_launch_userspace_program(char *command, int channel_no,
-		enum umh_wait wait);
+		enum umh_wait wait, int debug);
+
+extern char tuxonice_signature[9];
+extern int freezer_sync;
+
+extern int toi_start_other_threads(void);
+extern void toi_stop_other_threads(void);
 #endif
