@@ -1,29 +1,34 @@
 /*
  * kernel/power/tuxonice_alloc.c
  *
- * Copyright (C) 2007 Nigel Cunningham (nigel at tuxonice net)
+ * Copyright (C) 2008-2010 Nigel Cunningham (nigel at tuxonice net)
  *
  * This file is released under the GPLv2.
  *
  */
 
 #ifdef CONFIG_PM_DEBUG
-#include <linux/slab.h>
 #include <linux/module.h>
+#include <linux/slab.h>
 #include "tuxonice_modules.h"
+#include "tuxonice_alloc.h"
 #include "tuxonice_sysfs.h"
+#include "tuxonice.h"
 
-#define TOI_ALLOC_PATHS 38
+#define TOI_ALLOC_PATHS 40
 
-DEFINE_MUTEX(toi_alloc_mutex);
+static DEFINE_MUTEX(toi_alloc_mutex);
+
+static struct toi_module_ops toi_alloc_ops;
 
 static int toi_fail_num;
+
 static atomic_t toi_alloc_count[TOI_ALLOC_PATHS],
 		toi_free_count[TOI_ALLOC_PATHS],
 		toi_test_count[TOI_ALLOC_PATHS],
 		toi_fail_count[TOI_ALLOC_PATHS];
-int toi_cur_allocd[TOI_ALLOC_PATHS], toi_max_allocd[TOI_ALLOC_PATHS];
-int cur_allocd, max_allocd;
+static int toi_cur_allocd[TOI_ALLOC_PATHS], toi_max_allocd[TOI_ALLOC_PATHS];
+static int cur_allocd, max_allocd;
 
 static char *toi_alloc_desc[TOI_ALLOC_PATHS] = {
 	"", /* 0 */
@@ -63,7 +68,9 @@ static char *toi_alloc_desc[TOI_ALLOC_PATHS] = {
 	"sysfs kobj",
 	"swap mark resume attempted buffer", /* 35 */
 	"cluster member",
-	"boot kernel data buffer"
+	"boot kernel data buffer",
+	"setting swap signature",
+	"block i/o bdev struct"
 };
 
 #define MIGHT_FAIL(FAIL_NUM, FAIL_VAL) \
@@ -77,7 +84,7 @@ static char *toi_alloc_desc[TOI_ALLOC_PATHS] = {
 		} \
 	} while (0)
 
-static void alloc_update_stats(int fail_num, void *result)
+static void alloc_update_stats(int fail_num, void *result, int size)
 {
 	if (!result) {
 		atomic_inc(&toi_fail_count[fail_num]);
@@ -88,7 +95,7 @@ static void alloc_update_stats(int fail_num, void *result)
 	if (unlikely(test_action_state(TOI_GET_MAX_MEM_ALLOCD))) {
 		mutex_lock(&toi_alloc_mutex);
 		toi_cur_allocd[fail_num]++;
-		cur_allocd++;
+		cur_allocd += size;
 		if (unlikely(cur_allocd > max_allocd)) {
 			int i;
 
@@ -100,12 +107,16 @@ static void alloc_update_stats(int fail_num, void *result)
 	}
 }
 
-static void free_update_stats(int fail_num)
+static void free_update_stats(int fail_num, int size)
 {
+	BUG_ON(fail_num >= TOI_ALLOC_PATHS);
 	atomic_inc(&toi_free_count[fail_num]);
+	if (unlikely(atomic_read(&toi_free_count[fail_num]) >
+				atomic_read(&toi_alloc_count[fail_num])))
+		dump_stack();
 	if (unlikely(test_action_state(TOI_GET_MAX_MEM_ALLOCD))) {
 		mutex_lock(&toi_alloc_mutex);
-		cur_allocd--;
+		cur_allocd -= size;
 		toi_cur_allocd[fail_num]--;
 		mutex_unlock(&toi_alloc_mutex);
 	}
@@ -115,85 +126,123 @@ void *toi_kzalloc(int fail_num, size_t size, gfp_t flags)
 {
 	void *result;
 
-	MIGHT_FAIL(fail_num, NULL);
+	if (toi_alloc_ops.enabled)
+		MIGHT_FAIL(fail_num, NULL);
 	result = kzalloc(size, flags);
-	alloc_update_stats(fail_num, result);
+	if (toi_alloc_ops.enabled)
+		alloc_update_stats(fail_num, result, size);
+	if (fail_num == toi_trace_allocs)
+		dump_stack();
 	return result;
 }
+EXPORT_SYMBOL_GPL(toi_kzalloc);
 
 unsigned long toi_get_free_pages(int fail_num, gfp_t mask,
 		unsigned int order)
 {
 	unsigned long result;
 
-	MIGHT_FAIL(fail_num, 0);
+	if (toi_alloc_ops.enabled)
+		MIGHT_FAIL(fail_num, 0);
 	result = __get_free_pages(mask, order);
-	alloc_update_stats(fail_num, (void *) result);
+	if (toi_alloc_ops.enabled)
+		alloc_update_stats(fail_num, (void *) result,
+				PAGE_SIZE << order);
+	if (fail_num == toi_trace_allocs)
+		dump_stack();
 	return result;
 }
+EXPORT_SYMBOL_GPL(toi_get_free_pages);
 
 struct page *toi_alloc_page(int fail_num, gfp_t mask)
 {
 	struct page *result;
 
-	MIGHT_FAIL(fail_num, 0);
+	if (toi_alloc_ops.enabled)
+		MIGHT_FAIL(fail_num, NULL);
 	result = alloc_page(mask);
-	alloc_update_stats(fail_num, (void *) result);
+	if (toi_alloc_ops.enabled)
+		alloc_update_stats(fail_num, (void *) result, PAGE_SIZE);
+	if (fail_num == toi_trace_allocs)
+		dump_stack();
 	return result;
 }
+EXPORT_SYMBOL_GPL(toi_alloc_page);
 
 unsigned long toi_get_zeroed_page(int fail_num, gfp_t mask)
 {
 	unsigned long result;
 
-	MIGHT_FAIL(fail_num, 0);
+	if (toi_alloc_ops.enabled)
+		MIGHT_FAIL(fail_num, 0);
 	result = get_zeroed_page(mask);
-	alloc_update_stats(fail_num, (void *) result);
+	if (toi_alloc_ops.enabled)
+		alloc_update_stats(fail_num, (void *) result, PAGE_SIZE);
+	if (fail_num == toi_trace_allocs)
+		dump_stack();
 	return result;
 }
+EXPORT_SYMBOL_GPL(toi_get_zeroed_page);
 
-void toi_kfree(int fail_num, const void *arg)
+void toi_kfree(int fail_num, const void *arg, int size)
 {
-	if (arg)
-		free_update_stats(fail_num);
+	if (arg && toi_alloc_ops.enabled)
+		free_update_stats(fail_num, size);
 
+	if (fail_num == toi_trace_allocs)
+		dump_stack();
 	kfree(arg);
 }
+EXPORT_SYMBOL_GPL(toi_kfree);
 
 void toi_free_page(int fail_num, unsigned long virt)
 {
-	if (virt)
-		free_update_stats(fail_num);
+	if (virt && toi_alloc_ops.enabled)
+		free_update_stats(fail_num, PAGE_SIZE);
 
+	if (fail_num == toi_trace_allocs)
+		dump_stack();
 	free_page(virt);
 }
+EXPORT_SYMBOL_GPL(toi_free_page);
 
 void toi__free_page(int fail_num, struct page *page)
 {
-	if (page)
-		free_update_stats(fail_num);
+	if (page && toi_alloc_ops.enabled)
+		free_update_stats(fail_num, PAGE_SIZE);
 
+	if (fail_num == toi_trace_allocs)
+		dump_stack();
 	__free_page(page);
 }
+EXPORT_SYMBOL_GPL(toi__free_page);
 
 void toi_free_pages(int fail_num, struct page *page, int order)
 {
-	if (page)
-		free_update_stats(fail_num);
+	if (page && toi_alloc_ops.enabled)
+		free_update_stats(fail_num, PAGE_SIZE << order);
 
+	if (fail_num == toi_trace_allocs)
+		dump_stack();
 	__free_pages(page, order);
 }
 
 void toi_alloc_print_debug_stats(void)
 {
-	int i;
+	int i, header_done = 0;
 
-	printk(KERN_INFO "Idx  Allocs   Frees   Tests   Fails Max     "
-			"Description\n");
+	if (!toi_alloc_ops.enabled)
+		return;
 
 	for (i = 0; i < TOI_ALLOC_PATHS; i++)
-		if (atomic_read(&toi_alloc_count[i]) ||
-		    atomic_read(&toi_free_count[i]))
+		if (atomic_read(&toi_alloc_count[i]) !=
+		    atomic_read(&toi_free_count[i])) {
+			if (!header_done) {
+				printk(KERN_INFO "Idx  Allocs   Frees   Tests "
+					"  Fails     Max Description\n");
+				header_done = 1;
+			}
+
 			printk(KERN_INFO "%3d %7d %7d %7d %7d %7d %s\n", i,
 				atomic_read(&toi_alloc_count[i]),
 				atomic_read(&toi_free_count[i]),
@@ -201,6 +250,7 @@ void toi_alloc_print_debug_stats(void)
 				atomic_read(&toi_fail_count[i]),
 				toi_max_allocd[i],
 				toi_alloc_desc[i]);
+		}
 }
 EXPORT_SYMBOL_GPL(toi_alloc_print_debug_stats);
 
@@ -208,30 +258,34 @@ static int toi_alloc_initialise(int starting_cycle)
 {
 	int i;
 
-	if (starting_cycle) {
-		for (i = 0; i < TOI_ALLOC_PATHS; i++) {
-			atomic_set(&toi_alloc_count[i], 0);
-			atomic_set(&toi_free_count[i], 0);
-			atomic_set(&toi_test_count[i], 0);
-			atomic_set(&toi_fail_count[i], 0);
-			toi_cur_allocd[i] = 0;
-			toi_max_allocd[i] = 0;
-		};
-		max_allocd = 0;
-		cur_allocd = 0;
-	}
+	if (!starting_cycle)
+		return 0;
 
+	if (toi_trace_allocs)
+		dump_stack();
+
+	for (i = 0; i < TOI_ALLOC_PATHS; i++) {
+		atomic_set(&toi_alloc_count[i], 0);
+		atomic_set(&toi_free_count[i], 0);
+		atomic_set(&toi_test_count[i], 0);
+		atomic_set(&toi_fail_count[i], 0);
+		toi_cur_allocd[i] = 0;
+		toi_max_allocd[i] = 0;
+	};
+
+	max_allocd = 0;
+	cur_allocd = 0;
 	return 0;
 }
 
 static struct toi_sysfs_data sysfs_params[] = {
-	{ TOI_ATTR("failure_test", SYSFS_RW),
-	  SYSFS_INT(&toi_fail_num, 0, 99, 0)
-	},
-
-	{ TOI_ATTR("find_max_mem_allocated", SYSFS_RW),
-	  SYSFS_BIT(&toi_bkd.toi_action, TOI_GET_MAX_MEM_ALLOCD, 0)
-	}
+	SYSFS_INT("failure_test", SYSFS_RW, &toi_fail_num, 0, 99, 0, NULL),
+	SYSFS_INT("trace", SYSFS_RW, &toi_trace_allocs, 0, TOI_ALLOC_PATHS, 0,
+			NULL),
+	SYSFS_BIT("find_max_mem_allocated", SYSFS_RW, &toi_bkd.toi_action,
+			TOI_GET_MAX_MEM_ALLOCD, 0),
+	SYSFS_INT("enabled", SYSFS_RW, &toi_alloc_ops.enabled, 0, 1, 0,
+			NULL)
 };
 
 static struct toi_module_ops toi_alloc_ops = {
@@ -249,19 +303,12 @@ static struct toi_module_ops toi_alloc_ops = {
 
 int toi_alloc_init(void)
 {
-	return toi_register_module(&toi_alloc_ops);
+	int result = toi_register_module(&toi_alloc_ops);
+	return result;
 }
 
 void toi_alloc_exit(void)
 {
 	toi_unregister_module(&toi_alloc_ops);
 }
-#ifdef CONFIG_TOI_EXPORTS
-EXPORT_SYMBOL_GPL(toi_kzalloc);
-EXPORT_SYMBOL_GPL(toi_get_free_pages);
-EXPORT_SYMBOL_GPL(toi_get_zeroed_page);
-EXPORT_SYMBOL_GPL(toi_kfree);
-EXPORT_SYMBOL_GPL(toi_free_page);
-EXPORT_SYMBOL_GPL(toi__free_page);
-#endif
 #endif
