@@ -1,8 +1,7 @@
 /*
  * kernel/power/tuxonice_checksum.c
  *
- * Copyright (C) 2006-2007 Nigel Cunningham (nigel at tuxonice net)
- * Copyright (C) 2006 Red Hat, inc.
+ * Copyright (C) 2006-2010 Nigel Cunningham (nigel at tuxonice net)
  *
  * This file is released under the GPLv2.
  *
@@ -12,7 +11,6 @@
  */
 
 #include <linux/suspend.h>
-#include <linux/module.h>
 #include <linux/highmem.h>
 #include <linux/vmalloc.h>
 #include <linux/crypto.h>
@@ -26,6 +24,7 @@
 #include "tuxonice_checksum.h"
 #include "tuxonice_pagedir.h"
 #include "tuxonice_alloc.h"
+#include "tuxonice_ui.h"
 
 static struct toi_module_ops toi_checksum_ops;
 
@@ -124,7 +123,7 @@ static int toi_checksum_initialise(int starting_cycle)
 		if (!page)
 			return 1;
 		this->buf = page_address(page);
-		sg_set_buf(&this->sg[0], this->buf, PAGE_SIZE);
+		sg_init_one(&this->sg[0], this->buf, PAGE_SIZE);
 	}
 	return 0;
 }
@@ -143,12 +142,12 @@ static int toi_checksum_print_debug_stats(char *buffer, int size)
 	int len;
 
 	if (!toi_checksum_ops.enabled)
-		return snprintf_used(buffer, size,
+		return scnprintf(buffer, size,
 			"- Checksumming disabled.\n");
 
-	len = snprintf_used(buffer, size, "- Checksum method is '%s'.\n",
+	len = scnprintf(buffer, size, "- Checksum method is '%s'.\n",
 			toi_checksum_name);
-	len += snprintf_used(buffer + len, size - len,
+	len += scnprintf(buffer + len, size - len,
 		"  %d pages resaved in atomic copy.\n", toi_num_resaved);
 	return len;
 }
@@ -232,7 +231,7 @@ int allocate_checksum_pages(void)
 		unsigned long *new_page =
 		  (unsigned long *) toi_get_zeroed_page(15, TOI_ATOMIC_GFP);
 		if (!new_page) {
-			printk("Unable to allocate checksum pages.\n");
+			printk(KERN_ERR "Unable to allocate checksum pages.\n");
 			return -ENOMEM;
 		}
 		SetPageNosave(virt_to_page(new_page));
@@ -247,18 +246,6 @@ int allocate_checksum_pages(void)
 	return 0;
 }
 
-#if 0
-static void print_checksum(char *buf, int size)
-{
-	int index;
-
-	for (index = 0; index < size; index++)
-		printk(KERN_INFO "%x ", buf[index]);
-
-	printk("\n");
-}
-#endif
-
 char *tuxonice_get_next_checksum(void)
 {
 	if (!toi_checksum_ops.enabled)
@@ -289,6 +276,9 @@ int tuxonice_calc_checksum(struct page *page, char *checksum_locn)
 	kunmap(page);
 	result = crypto_hash_digest(&ctx->desc, ctx->sg, PAGE_SIZE,
 						checksum_locn);
+	if (result)
+		printk(KERN_ERR "TuxOnIce checksumming: crypto_hash_digest "
+				"returned %d.\n", result);
 	return result;
 }
 /*
@@ -298,18 +288,23 @@ int tuxonice_calc_checksum(struct page *page, char *checksum_locn)
 void check_checksums(void)
 {
 	int pfn, index = 0, cpu = smp_processor_id();
-	unsigned long next_page, this_checksum = 0;
 	char current_checksum[CHECKSUM_SIZE];
 	struct cpu_context *ctx = &per_cpu(contexts, cpu);
 
-	if (!toi_checksum_ops.enabled)
+	if (!toi_checksum_ops.enabled) {
+		toi_message(TOI_IO, TOI_VERBOSE, 0, "Checksumming disabled.");
 		return;
+	}
 
 	next_page = (unsigned long) page_list;
 
 	toi_num_resaved = 0;
+	this_checksum = 0;
 
-	BITMAP_FOR_EACH_SET(&pageset2_map, pfn) {
+	toi_message(TOI_IO, TOI_VERBOSE, 0, "Verifying checksums.");
+	memory_bm_position_reset(pageset2_map);
+	for (pfn = memory_bm_next_pfn(pageset2_map); pfn != BM_END_OF_MAP;
+			pfn = memory_bm_next_pfn(pageset2_map)) {
 		int ret;
 		char *pa;
 		struct page *page = pfn_to_page(pfn);
@@ -335,6 +330,8 @@ void check_checksums(void)
 
 		if (memcmp(current_checksum, (char *) this_checksum,
 							CHECKSUM_SIZE)) {
+			toi_message(TOI_IO, TOI_VERBOSE, 0, "Resaving %ld.",
+					pfn);
 			SetPageResave(pfn_to_page(pfn));
 			toi_num_resaved++;
 			if (test_action_state(TOI_ABORT_ON_RESAVE_NEEDED))
@@ -343,16 +340,14 @@ void check_checksums(void)
 
 		index++;
 	}
+	toi_message(TOI_IO, TOI_VERBOSE, 0, "Checksum verification complete.");
 }
 
 static struct toi_sysfs_data sysfs_params[] = {
-	{ TOI_ATTR("enabled", SYSFS_RW),
-	  SYSFS_INT(&toi_checksum_ops.enabled, 0, 1, 0)
-	},
-
-	{ TOI_ATTR("abort_if_resave_needed", SYSFS_RW),
-	  SYSFS_BIT(&toi_bkd.toi_action, TOI_ABORT_ON_RESAVE_NEEDED, 0)
-	}
+	SYSFS_INT("enabled", SYSFS_RW, &toi_checksum_ops.enabled, 0, 1, 0,
+			NULL),
+	SYSFS_BIT("abort_if_resave_needed", SYSFS_RW, &toi_bkd.toi_action,
+			TOI_ABORT_ON_RESAVE_NEEDED, 0)
 };
 
 /*
