@@ -1,7 +1,7 @@
 /*
  * kernel/power/compression.c
  *
- * Copyright (C) 2003-2007 Nigel Cunningham (nigel at tuxonice net)
+ * Copyright (C) 2003-2010 Nigel Cunningham (nigel at tuxonice net)
  *
  * This file is released under the GPLv2.
  *
@@ -9,11 +9,11 @@
  * using cryptoapi.
  */
 
-#include <linux/module.h>
 #include <linux/suspend.h>
 #include <linux/highmem.h>
 #include <linux/vmalloc.h>
 #include <linux/crypto.h>
+#include <linux/tuxonice.h>
 
 #include "tuxonice_builtin.h"
 #include "tuxonice.h"
@@ -28,7 +28,7 @@ static int toi_expected_compression;
 static struct toi_module_ops toi_compression_ops;
 static struct toi_module_ops *next_driver;
 
-static char toi_compressor_name[32] = "lzf";
+static char toi_compressor_name[32] = "lzo";
 
 static DEFINE_MUTEX(stats_lock);
 
@@ -36,39 +36,12 @@ struct cpu_context {
 	u8 *page_buffer;
 	struct crypto_comp *transform;
 	unsigned int len;
-	char *buffer_start;
+	u8 *buffer_start;
+	u8 *output_buffer;
 };
 
 static DEFINE_PER_CPU(struct cpu_context, contexts);
 
-static int toi_compress_prepare_result;
-
-/*
- * toi_compress_cleanup
- *
- * Frees memory allocated for our labours.
- */
-static void toi_compress_cleanup(int toi_or_resume)
-{
-	int cpu;
-
-	if (!toi_or_resume)
-		return;
-
-	for_each_online_cpu(cpu) {
-		struct cpu_context *this = &per_cpu(contexts, cpu);
-		if (this->transform) {
-			crypto_free_comp(this->transform);
-			this->transform = NULL;
-		}
-
-		if (this->page_buffer)
-			toi_free_page(16, (unsigned long) this->page_buffer);
-
-		this->page_buffer = NULL;
-	}
-}
-
 /*
  * toi_crypto_prepare
  *
@@ -101,9 +74,44 @@ static int toi_compress_crypto_prepare(void)
 		if (!this->page_buffer) {
 			printk(KERN_ERR
 			  "Failed to allocate a page buffer for TuxOnIce "
-			  "encryption driver.\n");
+			  "compression driver.\n");
 			return -ENOMEM;
 		}
+
+		this->output_buffer =
+			(char *) vmalloc_32(2 * PAGE_SIZE);
+
+		if (!this->output_buffer) {
+			printk(KERN_ERR
+			  "Failed to allocate a output buffer for TuxOnIce "
+			  "compression driver.\n");
+			return -ENOMEM;
+		}
+	}
+
+	return 0;
+}
+
+static int toi_compress_rw_cleanup(int writing)
+{
+	int cpu;
+
+	for_each_online_cpu(cpu) {
+		struct cpu_context *this = &per_cpu(contexts, cpu);
+		if (this->transform) {
+			crypto_free_comp(this->transform);
+			this->transform = NULL;
+		}
+
+		if (this->page_buffer)
+			toi_free_page(16, (unsigned long) this->page_buffer);
+
+		this->page_buffer = NULL;
+
+		if (this->output_buffer)
+			vfree(this->output_buffer);
+
+		this->output_buffer = NULL;
 	}
 
 	return 0;
@@ -118,30 +126,31 @@ static int toi_compress_init(int toi_or_resume)
 	if (!toi_or_resume)
 		return 0;
 
-	toi_compress_bytes_in = toi_compress_bytes_out = 0;
+	toi_compress_bytes_in = 0;
+	toi_compress_bytes_out = 0;
 
 	next_driver = toi_get_next_filter(&toi_compression_ops);
 
-	if (!next_driver)
-		return -ECHILD;
-
-	toi_compress_prepare_result = toi_compress_crypto_prepare();
-
-	return 0;
+	return next_driver ? 0 : -ECHILD;
 }
 
 /*
  * toi_compress_rw_init()
  */
 
-int toi_compress_rw_init(int rw, int stream_number)
+static int toi_compress_rw_init(int rw, int stream_number)
 {
-	if (toi_compress_prepare_result) {
-		printk("Failed to initialise compression algorithm.\n");
-		if (rw == READ)
+	if (toi_compress_crypto_prepare()) {
+		printk(KERN_ERR "Failed to initialise compression "
+				"algorithm.\n");
+		if (rw == READ) {
+			printk(KERN_INFO "Unable to read the image.\n");
 			return -ENODEV;
-		else
+		} else {
+			printk(KERN_INFO "Continuing without "
+				"compressing the image.\n");
 			toi_compression_ops.enabled = 0;
+		}
 	}
 
 	return 0;
@@ -160,44 +169,41 @@ int toi_compress_rw_init(int rw, int stream_number)
  * 		modules, -ECHILD if we have a broken pipeline or -EIO if
  * 		zlib errs.
  */
-static int toi_compress_write_page(unsigned long index,
-		struct page *buffer_page, unsigned int buf_size)
+static int toi_compress_write_page(unsigned long index, int buf_type,
+		void *buffer_page, unsigned int buf_size)
 {
 	int ret, cpu = smp_processor_id();
 	struct cpu_context *ctx = &per_cpu(contexts, cpu);
 
-	if (!ctx->transform)
-		return next_driver->write_page(index, buffer_page, buf_size);
+	if (!ctx->transform || PagePrecompressed((struct page *) buffer_page))
+		return next_driver->write_page(index, TOI_PAGE, buffer_page,
+				buf_size);
 
-	ctx->buffer_start = kmap(buffer_page);
+	ctx->buffer_start = TOI_MAP(buf_type, buffer_page);
 
-	ctx->len = buf_size;
+	ctx->len = PAGE_SIZE;
 
 	ret = crypto_comp_compress(ctx->transform,
 			ctx->buffer_start, buf_size,
-			ctx->page_buffer, &ctx->len);
+			ctx->output_buffer, &ctx->len);
 
-	kunmap(buffer_page);
-
-	if (ret) {
-		printk(KERN_INFO "Compression failed.\n");
-		goto failure;
-	}
+	TOI_UNMAP(buf_type, buffer_page);
 
 	mutex_lock(&stats_lock);
 	toi_compress_bytes_in += buf_size;
 	toi_compress_bytes_out += ctx->len;
 	mutex_unlock(&stats_lock);
 
-	if (ctx->len < buf_size) /* some compression */
-		ret = next_driver->write_page(index,
-				virt_to_page(ctx->page_buffer),
-				ctx->len);
-	else
-		ret = next_driver->write_page(index, buffer_page, buf_size);
+	toi_message(TOI_COMPRESS, TOI_VERBOSE, 0,
+			"CPU %d, index %lu: %d bytes",
+			cpu, index, ctx->len);
 
-failure:
-	return ret;
+	if (!ret && ctx->len < buf_size) /* some compression */
+		return next_driver->write_page(index, TOI_VIRT,
+				ctx->output_buffer, ctx->len);
+	else
+		return next_driver->write_page(index, TOI_PAGE, buffer_page,
+				buf_size);
 }
 
 /*
@@ -208,8 +214,8 @@ failure:
  * is filled.
  * Zero if successful. Error condition from me or from downstream on failure.
  */
-static int toi_compress_read_page(unsigned long *index,
-		struct page *buffer_page, unsigned int *buf_size)
+static int toi_compress_read_page(unsigned long *index, int buf_type,
+		void *buffer_page, unsigned int *buf_size)
 {
 	int ret, cpu = smp_processor_id();
 	unsigned int len;
@@ -218,27 +224,33 @@ static int toi_compress_read_page(unsigned long *index,
 	struct cpu_context *ctx = &per_cpu(contexts, cpu);
 
 	if (!ctx->transform)
-		return next_driver->read_page(index, buffer_page, buf_size);
+		return next_driver->read_page(index, TOI_PAGE, buffer_page,
+				buf_size);
 
 	/*
 	 * All our reads must be synchronous - we can't decompress
 	 * data that hasn't been read yet.
 	 */
 
-	*buf_size = PAGE_SIZE;
-
-	ret = next_driver->read_page(index, buffer_page, &len);
-
-	/* Error or uncompressed data */
-	if (ret || len == PAGE_SIZE)
-		return ret;
+	ret = next_driver->read_page(index, TOI_VIRT, ctx->page_buffer, &len);
 
 	buffer_start = kmap(buffer_page);
-	memcpy(ctx->page_buffer, buffer_start, len);
+
+	/* Error or uncompressed data */
+	if (ret || len == PAGE_SIZE) {
+		memcpy(buffer_start, ctx->page_buffer, len);
+		goto out;
+	}
+
 	ret = crypto_comp_decompress(
 			ctx->transform,
 			ctx->page_buffer,
 			len, buffer_start, &outlen);
+
+	toi_message(TOI_COMPRESS, TOI_VERBOSE, 0,
+			"CPU %d, index %lu: %d=>%d (%d).",
+			cpu, *index, len, outlen, ret);
+
 	if (ret)
 		abort_hibernate(TOI_FAILED_IO,
 			"Compress_read returned %d.\n", ret);
@@ -246,10 +258,13 @@ static int toi_compress_read_page(unsigned long *index,
 		abort_hibernate(TOI_FAILED_IO,
 			"Decompression yielded %d bytes instead of %ld.\n",
 			outlen, PAGE_SIZE);
+		printk(KERN_ERR "Decompression yielded %d bytes instead of "
+				"%ld.\n", outlen, PAGE_SIZE);
 		ret = -EIO;
 		*buf_size = outlen;
 	}
-	kunmap(buffer_page);
+out:
+	TOI_UNMAP(buf_type, buffer_page);
 	return ret;
 }
 
@@ -270,14 +285,14 @@ static int toi_compress_print_debug_stats(char *buffer, int size)
 
 	/* Output the compression ratio achieved. */
 	if (*toi_compressor_name)
-		len = snprintf_used(buffer, size, "- Compressor is '%s'.\n",
+		len = scnprintf(buffer, size, "- Compressor is '%s'.\n",
 				toi_compressor_name);
 	else
-		len = snprintf_used(buffer, size, "- Compressor is not set.\n");
+		len = scnprintf(buffer, size, "- Compressor is not set.\n");
 
 	if (pages_in)
-		len += snprintf_used(buffer+len, size - len,
-		  "  Compressed %lu bytes into %lu (%d percent compression).\n",
+		len += scnprintf(buffer+len, size - len, "  Compressed "
+			"%lu bytes into %lu (%ld percent compression).\n",
 		  toi_compress_bytes_in,
 		  toi_compress_bytes_out,
 		  (pages_in - pages_out) * 100 / pages_in);
@@ -298,7 +313,8 @@ static int toi_compress_memory_needed(void)
 
 static int toi_compress_storage_needed(void)
 {
-	return 4 * sizeof(unsigned long) + strlen(toi_compressor_name) + 1;
+	return 2 * sizeof(unsigned long) + 2 * sizeof(int) +
+		strlen(toi_compressor_name) + 1;
 }
 
 /*
@@ -310,19 +326,18 @@ static int toi_compress_storage_needed(void)
  */
 static int toi_compress_save_config_info(char *buffer)
 {
-	int namelen = strlen(toi_compressor_name) + 1;
-	int total_len;
+	int len = strlen(toi_compressor_name) + 1, offset = 0;
 
 	*((unsigned long *) buffer) = toi_compress_bytes_in;
-	*((unsigned long *) (buffer + 1 * sizeof(unsigned long))) =
-		toi_compress_bytes_out;
-	*((unsigned long *) (buffer + 2 * sizeof(unsigned long))) =
-		toi_expected_compression;
-	*((unsigned long *) (buffer + 3 * sizeof(unsigned long))) = namelen;
-	strncpy(buffer + 4 * sizeof(unsigned long), toi_compressor_name,
-								namelen);
-	total_len = 4 * sizeof(unsigned long) + namelen;
-	return total_len;
+	offset += sizeof(unsigned long);
+	*((unsigned long *) (buffer + offset)) = toi_compress_bytes_out;
+	offset += sizeof(unsigned long);
+	*((int *) (buffer + offset)) = toi_expected_compression;
+	offset += sizeof(int);
+	*((int *) (buffer + offset)) = len;
+	offset += sizeof(int);
+	strncpy(buffer + offset, toi_compressor_name, len);
+	return offset + len;
 }
 
 /* toi_compress_load_config_info
@@ -334,17 +349,29 @@ static int toi_compress_save_config_info(char *buffer)
  */
 static void toi_compress_load_config_info(char *buffer, int size)
 {
-	int namelen;
+	int len, offset = 0;
 
 	toi_compress_bytes_in = *((unsigned long *) buffer);
-	toi_compress_bytes_out = *((unsigned long *) (buffer + 1 *
-				sizeof(unsigned long)));
-	toi_expected_compression = *((unsigned long *) (buffer + 2 *
-				sizeof(unsigned long)));
-	namelen = *((unsigned long *) (buffer + 3 * sizeof(unsigned long)));
-	strncpy(toi_compressor_name, buffer + 4 * sizeof(unsigned long),
-			namelen);
-	return;
+	offset += sizeof(unsigned long);
+	toi_compress_bytes_out = *((unsigned long *) (buffer + offset));
+	offset += sizeof(unsigned long);
+	toi_expected_compression = *((int *) (buffer + offset));
+	offset += sizeof(int);
+	len = *((int *) (buffer + offset));
+	offset += sizeof(int);
+	strncpy(toi_compressor_name, buffer + offset, len);
+}
+
+static void toi_compress_pre_atomic_restore(struct toi_boot_kernel_data *bkd)
+{
+	bkd->compress_bytes_in = toi_compress_bytes_in;
+	bkd->compress_bytes_out = toi_compress_bytes_out;
+}
+
+static void toi_compress_post_atomic_restore(struct toi_boot_kernel_data *bkd)
+{
+	toi_compress_bytes_in = bkd->compress_bytes_in;
+	toi_compress_bytes_out = bkd->compress_bytes_out;
 }
 
 /*
@@ -368,20 +395,11 @@ static int toi_compress_expected_ratio(void)
  * data for our sysfs entries.
  */
 static struct toi_sysfs_data sysfs_params[] = {
-	{
-		TOI_ATTR("expected_compression", SYSFS_RW),
-		SYSFS_INT(&toi_expected_compression, 0, 99, 0)
-	},
-
-	{
-		TOI_ATTR("enabled", SYSFS_RW),
-		SYSFS_INT(&toi_compression_ops.enabled, 0, 1, 0)
-	},
-
-	{
-		TOI_ATTR("algorithm", SYSFS_RW),
-		SYSFS_STRING(toi_compressor_name, 31, 0)
-	}
+	SYSFS_INT("expected_compression", SYSFS_RW, &toi_expected_compression,
+			0, 99, 0, NULL),
+	SYSFS_INT("enabled", SYSFS_RW, &toi_compression_ops.enabled, 0, 1, 0,
+			NULL),
+	SYSFS_STRING("algorithm", SYSFS_RW, toi_compressor_name, 31, 0, NULL),
 };
 
 /*
@@ -393,7 +411,6 @@ static struct toi_module_ops toi_compression_ops = {
 	.directory		= "compression",
 	.module			= THIS_MODULE,
 	.initialise		= toi_compress_init,
-	.cleanup		= toi_compress_cleanup,
 	.memory_needed 		= toi_compress_memory_needed,
 	.print_debug_info	= toi_compress_print_debug_stats,
 	.save_config_info	= toi_compress_save_config_info,
@@ -401,7 +418,11 @@ static struct toi_module_ops toi_compression_ops = {
 	.storage_needed		= toi_compress_storage_needed,
 	.expected_compression	= toi_compress_expected_ratio,
 
+	.pre_atomic_restore	= toi_compress_pre_atomic_restore,
+	.post_atomic_restore	= toi_compress_post_atomic_restore,
+
 	.rw_init		= toi_compress_rw_init,
+	.rw_cleanup		= toi_compress_rw_cleanup,
 
 	.write_page		= toi_compress_write_page,
 	.read_page		= toi_compress_read_page,
