@@ -1,7 +1,7 @@
 /*
  * kernel/power/tuxonice_modules.h
  *
- * Copyright (C) 2004-2007 Nigel Cunningham (nigel at tuxonice net)
+ * Copyright (C) 2004-2010 Nigel Cunningham (nigel at tuxonice net)
  *
  * This file is released under the GPLv2.
  *
@@ -17,6 +17,8 @@
 /* This is the maximum size we store in the image header for a module name */
 #define TOI_MAX_MODULE_NAME_LENGTH 30
 
+struct toi_boot_kernel_data;
+
 /* Per-module metadata */
 struct toi_module_header {
 	char name[TOI_MAX_MODULE_NAME_LENGTH];
@@ -30,7 +32,8 @@ struct toi_module_header {
 enum {
 	FILTER_MODULE,
 	WRITER_MODULE,
-	MISC_MODULE, /* Block writer, eg. */
+	BIO_ALLOCATOR_MODULE,
+	MISC_MODULE,
 	MISC_HIDDEN_MODULE,
 };
 
@@ -39,6 +42,20 @@ enum {
 	TOI_SYNC
 };
 
+enum {
+	TOI_VIRT,
+	TOI_PAGE,
+};
+
+#define TOI_MAP(type, addr) \
+ (type == TOI_PAGE ? kmap(addr) : addr)
+
+#define TOI_UNMAP(type, addr) \
+ do { \
+   if (type == TOI_PAGE) \
+     kunmap(addr); \
+ } while(0)
+
 struct toi_module_ops {
 	/* Functions common to all modules */
 	int type;
@@ -47,7 +64,7 @@ struct toi_module_ops {
 	char *shared_directory;
 	struct kobject *dir_kobj;
 	struct module *module;
-	int enabled, early;
+	int enabled, early, initialised;
 	struct list_head module_list;
 
 	/* List of filters or allocators */
@@ -78,32 +95,35 @@ struct toi_module_ops {
 	int (*initialise) (int starting_cycle);
 	void (*cleanup) (int finishing_cycle);
 
+	void (*pre_atomic_restore) (struct toi_boot_kernel_data *bkd);
+	void (*post_atomic_restore) (struct toi_boot_kernel_data *bkd);
+
 	/*
 	 * Calls for allocating storage (allocators only).
 	 *
-	 * Header space is allocated separately. Note that allocation
-	 * of space for the header might result in allocated space
-	 * being stolen from the main pool if there is no unallocated
-	 * space. We have to be able to allocate enough space for
-	 * the header. We can eat memory to ensure there is enough
-	 * for the main pool.
+	 * Header space is requested separately and cannot fail, but the
+	 * reservation is only applied when main storage is allocated.
+	 * The header space reservation is thus always set prior to
+	 * requesting the allocation of storage - and prior to querying
+	 * how much storage is available.
 	 */
 
-	int (*storage_available) (void);
-	int (*allocate_header_space) (int space_requested);
-	int (*allocate_storage) (int space_requested);
-	int (*storage_allocated) (void);
-	int (*release_storage) (void);
+	unsigned long (*storage_available) (void);
+	void (*reserve_header_space) (unsigned long space_requested);
+	int (*register_storage) (void);
+	int (*allocate_storage) (unsigned long space_requested);
+	unsigned long (*storage_allocated) (void);
 
 	/*
 	 * Routines used in image I/O.
 	 */
 	int (*rw_init) (int rw, int stream_number);
 	int (*rw_cleanup) (int rw);
-	int (*write_page) (unsigned long index, struct page *buffer_page,
+	int (*write_page) (unsigned long index, int buf_type, void *buf,
 			unsigned int buf_size);
-	int (*read_page) (unsigned long *index, struct page *buffer_page,
+	int (*read_page) (unsigned long *index, int buf_type, void *buf,
 			unsigned int *buf_size);
+	int (*io_flusher) (int rw);
 
 	/* Reset module if image exists but reading aborted */
 	void (*noresume_reset) (void);
@@ -115,17 +135,30 @@ struct toi_module_ops {
 	int (*read_header_init) (void);
 	int (*read_header_cleanup) (void);
 
+	/* To be called after read_header_init */
+	int (*get_header_version) (void);
+
 	int (*rw_header_chunk) (int rw, struct toi_module_ops *owner,
 			char *buffer_start, int buffer_size);
 
+	int (*rw_header_chunk_noreadahead) (int rw,
+			struct toi_module_ops *owner, char *buffer_start,
+			int buffer_size);
+
 	/* Attempt to parse an image location */
 	int (*parse_sig_location) (char *buffer, int only_writer, int quiet);
 
+	/* Throttle I/O according to throughput */
+	void (*update_throughput_throttle) (int jif_index);
+
+	/* Flush outstanding I/O */
+	int (*finish_all_io) (void);
+
 	/* Determine whether image exists that we can restore */
-	int (*image_exists) (void);
+	int (*image_exists) (int quiet);
 
 	/* Mark the image as having tried to resume */
-	void (*mark_resume_attempted) (int);
+	int (*mark_resume_attempted) (int);
 
 	/* Destroy image if one exists */
 	int (*remove_image) (void);
@@ -133,6 +166,9 @@ struct toi_module_ops {
 	/* Sysfs Data */
 	struct toi_sysfs_data *sysfs_data;
 	int num_sysfs_entries;
+
+	/* Block I/O allocator */
+	struct toi_bio_allocator_ops *bio_allocator_ops;
 };
 
 extern int toi_num_modules, toiNumAllocators;
@@ -149,8 +185,9 @@ extern struct toi_module_ops *toi_get_next_filter(struct toi_module_ops *);
 extern int toi_register_module(struct toi_module_ops *module);
 extern void toi_move_module_tail(struct toi_module_ops *module);
 
-extern int toi_header_storage_for_modules(void);
-extern int toi_memory_for_modules(int print_parts);
+extern long toi_header_storage_for_modules(void);
+extern long toi_memory_for_modules(int print_parts);
+extern void print_toi_header_storage_for_modules(void);
 extern int toi_expected_compression_ratio(void);
 
 extern int toi_print_module_debug_info(char *buffer, int buffer_size);
@@ -164,6 +201,9 @@ extern int toi_initialise_modules(int starting_cycle, int early);
 	toi_initialise_modules(starting, 0)
 extern void toi_cleanup_modules(int finishing_cycle);
 
+extern void toi_post_atomic_restore_modules(struct toi_boot_kernel_data *bkd);
+extern void toi_pre_atomic_restore_modules(struct toi_boot_kernel_data *bkd);
+
 extern void toi_print_modules(void);
 
 int toi_get_modules(void);
