@@ -1,7 +1,7 @@
 /*
  * kernel/power/tuxonice_netlink.h
  *
- * Copyright (C) 2004-2007 Nigel Cunningham (nigel at tuxonice net)
+ * Copyright (C) 2004-2010 Nigel Cunningham (nigel at tuxonice net)
  *
  * This file is released under the GPLv2.
  *
@@ -37,8 +37,9 @@ struct user_helper_data {
 	char *name;
 	struct user_helper_data *next;
 	struct completion wait_for_process;
-	int interface_version;
+	u32 interface_version;
 	int must_init;
+	int debug;
 };
 
 #ifdef CONFIG_NET
@@ -46,6 +47,7 @@ int toi_netlink_setup(struct user_helper_data *uhd);
 void toi_netlink_close(struct user_helper_data *uhd);
 void toi_send_netlink_message(struct user_helper_data *uhd,
 		int type, void *params, size_t len);
+void toi_netlink_close_complete(struct user_helper_data *uhd);
 #else
 static inline int toi_netlink_setup(struct user_helper_data *uhd)
 {
@@ -55,4 +57,6 @@ static inline int toi_netlink_setup(struct user_helper_data *uhd)
 static inline void toi_netlink_close(struct user_helper_data *uhd) { };
 static inline void toi_send_netlink_message(struct user_helper_data *uhd,
 		int type, void *params, size_t len) { };
+static inline void toi_netlink_close_complete(struct user_helper_data *uhd)
+	{ };
 #endif
