@@ -1,7 +1,7 @@
 /*
  * kernel/power/tuxonice_pagedir.h
  *
- * Copyright (C) 2006-2007 Nigel Cunningham (nigel at tuxonice net)
+ * Copyright (C) 2006-2010 Nigel Cunningham (nigel at tuxonice net)
  *
  * This file is released under the GPLv2.
  *
@@ -18,9 +18,9 @@
 
 struct pagedir {
 	int id;
-	int size;
+	unsigned long size;
 #ifdef CONFIG_HIGHMEM
-	int size_high;
+	unsigned long size_high;
 #endif
 };
 
