@@ -1,7 +1,7 @@
 /*
  * kernel/power/tuxonice_power_off.c
  *
- * Copyright (C) 2006-2007 Nigel Cunningham (nigel at tuxonice net)
+ * Copyright (C) 2006-2010 Nigel Cunningham (nigel at tuxonice net)
  *
  * This file is released under the GPLv2.
  *
@@ -21,22 +21,17 @@
 #include "tuxonice_power_off.h"
 #include "tuxonice_sysfs.h"
 #include "tuxonice_modules.h"
+#include "tuxonice_io.h"
 
 unsigned long toi_poweroff_method; /* 0 - Kernel power off */
 EXPORT_SYMBOL_GPL(toi_poweroff_method);
 
-int wake_delay;
+static int wake_delay;
 static char lid_state_file[256], wake_alarm_dir[256];
 static struct file *lid_file, *alarm_file, *epoch_file;
-int post_wake_state = -1;
+static int post_wake_state = -1;
 
-extern struct hibernation_ops *hibernation_ops;
-
-int toi_platform_prepare(void)
-{
-	return (toi_poweroff_method == 4 && hibernation_ops) ?
-		hibernation_ops->prepare() : 0;
-}
+static int did_suspend_to_both;
 
 /*
  * __toi_power_down
@@ -51,27 +46,52 @@ static void __toi_power_down(int method)
 {
 	int error;
 
-	if (test_action_state(TOI_REBOOT)) {
-		toi_prepare_status(DONT_CLEAR_BAR, "Ready to reboot.");
-		kernel_restart(NULL);
-	}
+	toi_cond_pause(1, test_action_state(TOI_REBOOT) ? "Ready to reboot." :
+			"Powering down.");
 
-	toi_prepare_status(DONT_CLEAR_BAR, "Powering down.");
+	if (test_result_state(TOI_ABORTED))
+		goto out;
+
+	if (test_action_state(TOI_REBOOT))
+		kernel_restart(NULL);
 
 	switch (method) {
 	case 0:
 		break;
 	case 3:
+		/*
+		 * Re-read the overwritten part of pageset2 to make post-resume
+		 * faster.
+		 */
+		if (read_pageset2(1))
+			panic("Attempt to reload pagedir 2 failed. "
+					"Try rebooting.");
+
+		pm_prepare_console();
+
 		error = pm_notifier_call_chain(PM_SUSPEND_PREPARE);
-		if (!error)
+		if (!error) {
+			pm_restore_gfp_mask();
 			error = suspend_devices_and_enter(PM_SUSPEND_MEM);
+			pm_restrict_gfp_mask();
+			if (!error)
+				did_suspend_to_both = 1;
+		}
 		pm_notifier_call_chain(PM_POST_SUSPEND);
-		if (!error)
+		pm_restore_console();
+
+		/* Success - we're now post-resume-from-ram */
+		if (did_suspend_to_both)
 			return;
+
+		/* Failed to suspend to ram - do normal power off */
 		break;
 	case 4:
-		if (!hibernation_platform_enter())
-			return;
+		/*
+		 * If succeeds, doesn't return. If fails, do a simple
+		 * powerdown.
+		 */
+		hibernation_platform_enter();
 		break;
 	case 5:
 		/* Historic entry only now */
@@ -79,23 +99,22 @@ static void __toi_power_down(int method)
 	}
 
 	if (method && method != 5)
-		toi_prepare_status(DONT_CLEAR_BAR,
-				"Falling back to alternate power off method.");
+		toi_cond_pause(1,
+			"Falling back to alternate power off method.");
 
 	if (test_result_state(TOI_ABORTED))
-		return;
+		goto out;
 
 	kernel_power_off();
 	kernel_halt();
-	toi_prepare_status(DONT_CLEAR_BAR, "Powerdown failed.");
+	toi_cond_pause(1, "Powerdown failed.");
 	while (1)
 		cpu_relax();
-}
 
-void toi_platform_finish(void)
-{
-	if (toi_poweroff_method == 4 && hibernation_ops)
-		hibernation_ops->finish();
+out:
+	if (read_pageset2(1))
+		panic("Attempt to reload pagedir 2 failed. Try rebooting.");
+	return;
 }
 
 #define CLOSE_FILE(file) \
@@ -103,7 +122,7 @@ void toi_platform_finish(void)
 		filp_close(file, NULL); file = NULL; \
 	}
 
-static void powerdown_files_close(int toi_or_resume)
+static void powerdown_cleanup(int toi_or_resume)
 {
 	if (!toi_or_resume)
 		return;
@@ -124,16 +143,18 @@ static void open_file(char *format, char *arg, struct file **var, int mode,
 		if (IS_ERR(*var) || !*var) {
 			printk(KERN_INFO "Failed to open %s file '%s' (%p).\n",
 				desc, buf, *var);
-			*var = 0;
+			*var = NULL;
 		}
 	}
 }
 
-static int powerdown_files_open(int toi_or_resume)
+static int powerdown_init(int toi_or_resume)
 {
 	if (!toi_or_resume)
 		return 0;
 
+	did_suspend_to_both = 0;
+
 	open_file("/proc/acpi/button/%s/state", lid_state_file, &lid_file,
 			O_RDONLY, "lid");
 
@@ -213,14 +234,14 @@ void toi_power_down(void)
 			printk(KERN_INFO "Failed to read epoch file (%d).\n",
 					(int) size);
 		else {
-			unsigned long since_epoch =
-				simple_strtol(array, NULL, 0);
+			unsigned long since_epoch;
+			if (!strict_strtoul(array, 0, &since_epoch)) {
+				/* Clear any wakeup time. */
+				write_alarm_file(0);
 
-			/* Clear any wakeup time. */
-			write_alarm_file(0);
-
-			/* Set new wakeup time. */
-			write_alarm_file(since_epoch + wake_delay);
+				/* Set new wakeup time. */
+				write_alarm_file(since_epoch + wake_delay);
+			}
 		}
 	}
 
@@ -232,36 +253,22 @@ EXPORT_SYMBOL_GPL(toi_power_down);
 
 static struct toi_sysfs_data sysfs_params[] = {
 #if defined(CONFIG_ACPI)
-	{
-	 TOI_ATTR("lid_file", SYSFS_RW),
-	 SYSFS_STRING(lid_state_file, 256, 0),
-	},
-
-	{
-	  TOI_ATTR("wake_delay", SYSFS_RW),
-	  SYSFS_INT(&wake_delay, 0, INT_MAX, 0)
-	},
-
-	{
-	  TOI_ATTR("wake_alarm_dir", SYSFS_RW),
-	  SYSFS_STRING(wake_alarm_dir, 256, 0)
-	},
-
-	{ TOI_ATTR("post_wake_state", SYSFS_RW),
-	  SYSFS_INT(&post_wake_state, -1, 5, 0)
-	},
-
-	{ TOI_ATTR("powerdown_method", SYSFS_RW),
-	  SYSFS_UL(&toi_poweroff_method, 0, 5, 0)
-	},
+	SYSFS_STRING("lid_file", SYSFS_RW, lid_state_file, 256, 0, NULL),
+	SYSFS_INT("wake_delay", SYSFS_RW, &wake_delay, 0, INT_MAX, 0, NULL),
+	SYSFS_STRING("wake_alarm_dir", SYSFS_RW, wake_alarm_dir, 256, 0, NULL),
+	SYSFS_INT("post_wake_state", SYSFS_RW, &post_wake_state, -1, 5, 0,
+			NULL),
+	SYSFS_UL("powerdown_method", SYSFS_RW, &toi_poweroff_method, 0, 5, 0),
+	SYSFS_INT("did_suspend_to_both", SYSFS_READONLY, &did_suspend_to_both,
+		0, 0, 0, NULL)
 #endif
 };
 
 static struct toi_module_ops powerdown_ops = {
 	.type				= MISC_HIDDEN_MODULE,
 	.name				= "poweroff",
-	.initialise			= powerdown_files_open,
-	.cleanup			= powerdown_files_close,
+	.initialise			= powerdown_init,
+	.cleanup			= powerdown_cleanup,
 	.directory			= "[ROOT]",
 	.module				= THIS_MODULE,
 	.sysfs_data			= sysfs_params,
