@@ -1,7 +1,7 @@
 /*
  * kernel/power/tuxonice_prepare_image.c
  *
- * Copyright (C) 2003-2007 Nigel Cunningham (nigel at tuxonice net)
+ * Copyright (C) 2003-2010 Nigel Cunningham (nigel at tuxonice net)
  *
  * This file is released under the GPLv2.
  *
@@ -20,7 +20,6 @@
  *
  */
 
-#include <linux/module.h>
 #include <linux/highmem.h>
 #include <linux/freezer.h>
 #include <linux/hardirq.h>
@@ -31,18 +30,21 @@
 #include "tuxonice_modules.h"
 #include "tuxonice_io.h"
 #include "tuxonice_ui.h"
-#include "tuxonice_extent.h"
 #include "tuxonice_prepare_image.h"
-#include "tuxonice_block_io.h"
 #include "tuxonice.h"
+#include "tuxonice_extent.h"
 #include "tuxonice_checksum.h"
 #include "tuxonice_sysfs.h"
 #include "tuxonice_alloc.h"
+#include "tuxonice_atomic_copy.h"
+#include "tuxonice_builtin.h"
 
-static int num_nosave, header_space_allocated, main_storage_allocated,
-	   storage_available;
-int extra_pd1_pages_allowance = MIN_EXTRA_PAGES_ALLOWANCE;
-int image_size_limit;
+static unsigned long num_nosave, main_storage_allocated, storage_limit,
+	    header_storage_needed;
+unsigned long extra_pd1_pages_allowance =
+	CONFIG_TOI_DEFAULT_EXTRA_PAGES_ALLOWANCE;
+long image_size_limit = CONFIG_TOI_DEFAULT_IMAGE_SIZE_LIMIT;
+static int no_ps2_needed;
 
 struct attention_list {
 	struct task_struct *task;
@@ -61,7 +63,7 @@ void free_attention_list(void)
 	while (attention_list) {
 		last = attention_list;
 		attention_list = attention_list->next;
-		toi_kfree(6, last);
+		toi_kfree(6, last, sizeof(*last));
 	}
 }
 
@@ -74,11 +76,11 @@ static int build_attention_list(void)
 	/*
 	 * Count all userspace process (with task->mm) marked PF_NOFREEZE.
 	 */
-	read_lock(&tasklist_lock);
+	toi_read_lock_tasklist();
 	for_each_process(p)
 		if ((p->flags & PF_NOFREEZE) || p == current)
 			task_count++;
-	read_unlock(&tasklist_lock);
+	toi_read_unlock_tasklist();
 
 	/*
 	 * Allocate attention list structs.
@@ -100,32 +102,37 @@ static int build_attention_list(void)
 	}
 
 	next = attention_list;
-	read_lock(&tasklist_lock);
+	toi_read_lock_tasklist();
 	for_each_process(p)
 		if ((p->flags & PF_NOFREEZE) || p == current) {
 			next->task = p;
 			next = next->next;
 		}
-	read_unlock(&tasklist_lock);
+	toi_read_unlock_tasklist();
 	return 0;
 }
 
 static void pageset2_full(void)
 {
 	struct zone *zone;
+	struct page *page;
 	unsigned long flags;
+	int i;
 
-	for_each_zone(zone) {
+	for_each_populated_zone(zone) {
 		spin_lock_irqsave(&zone->lru_lock, flags);
-		if (zone_page_state(zone, NR_INACTIVE)) {
-			struct page *page;
-			list_for_each_entry(page, &zone->inactive_list, lru)
-				SetPagePageset2(page);
-		}
-		if (zone_page_state(zone, NR_ACTIVE)) {
-			struct page *page;
-			list_for_each_entry(page, &zone->active_list, lru)
-				SetPagePageset2(page);
+		for_each_lru(i) {
+			if (!zone_page_state(zone, NR_LRU_BASE + i))
+				continue;
+
+			list_for_each_entry(page, &zone->lru[i].list, lru) {
+				struct address_space *mapping;
+
+				mapping = page_mapping(page);
+				if (!mapping || !mapping->host ||
+				    !(mapping->host->i_flags & S_ATOMIC_COPY))
+					SetPagePageset2(page);
+			}
 		}
 		spin_unlock_irqrestore(&zone->lru_lock, flags);
 	}
@@ -153,14 +160,21 @@ static void toi_mark_task_as_pageset(struct task_struct *t, int pageset2)
 	for (vma = mm->mmap; vma; vma = vma->vm_next) {
 		unsigned long posn;
 
-		if (vma->vm_flags & (VM_PFNMAP | VM_IO | VM_RESERVED) ||
-		    !vma->vm_start)
+		if (!vma->vm_start ||
+		    vma->vm_flags & (VM_IO | VM_RESERVED | VM_PFNMAP))
 			continue;
 
 		for (posn = vma->vm_start; posn < vma->vm_end;
 				posn += PAGE_SIZE) {
 			struct page *page = follow_page(vma, posn, 0);
-			if (!page)
+			struct address_space *mapping;
+
+			if (!page || !pfn_valid(page_to_pfn(page)))
+				continue;
+
+			mapping = page_mapping(page);
+			if (mapping && mapping->host &&
+			    mapping->host->i_flags & S_ATOMIC_COPY)
 				continue;
 
 			if (pageset2)
@@ -176,6 +190,24 @@ static void toi_mark_task_as_pageset(struct task_struct *t, int pageset2)
 		up_read(&mm->mmap_sem);
 }
 
+static void mark_tasks(int pageset)
+{
+	struct task_struct *p;
+
+	toi_read_lock_tasklist();
+	for_each_process(p) {
+		if (!p->mm)
+			continue;
+
+		if (p->flags & PF_KTHREAD)
+			continue;
+
+		toi_mark_task_as_pageset(p, pageset);
+	}
+	toi_read_unlock_tasklist();
+
+}
+
 /* mark_pages_for_pageset2
  *
  * Description:	Mark unshared pages in processes not needed for hibernate as
@@ -186,26 +218,17 @@ static void toi_mark_task_as_pageset(struct task_struct *t, int pageset2)
 
 static void toi_mark_pages_for_pageset2(void)
 {
-	struct task_struct *p;
 	struct attention_list *this = attention_list;
 
-	if (test_action_state(TOI_NO_PAGESET2))
-		return;
+	memory_bm_clear(pageset2_map);
 
-	clear_dyn_pageflags(&pageset2_map);
+	if (test_action_state(TOI_NO_PAGESET2) || no_ps2_needed)
+		return;
 
 	if (test_action_state(TOI_PAGESET2_FULL))
 		pageset2_full();
-	else {
-		read_lock(&tasklist_lock);
-		for_each_process(p) {
-			if (!p->mm || (p->flags & PF_BORROWED_MM))
-				continue;
-
-			toi_mark_task_as_pageset(p, PAGESET2);
-		}
-		read_unlock(&tasklist_lock);
-	}
+	else
+		mark_tasks(PAGESET2);
 
 	/*
 	 * Because the tasks in attention_list are ones related to hibernating,
@@ -227,7 +250,7 @@ static void toi_mark_pages_for_pageset2(void)
  * the allocation and freeing of that memory.
  */
 
-static int extra_pages_allocated;
+static unsigned long extra_pages_allocated;
 
 struct extras {
 	struct page *page;
@@ -254,7 +277,7 @@ void toi_free_extra_pagedir_memory(void)
 			ClearPageNosave(this->page + i);
 
 		toi_free_pages(9, this->page, this->order);
-		toi_kfree(7, this);
+		toi_kfree(7, this, sizeof(*this));
 	}
 
 	extra_pages_allocated = 0;
@@ -270,7 +293,7 @@ void toi_free_extra_pagedir_memory(void)
 static int toi_allocate_extra_pagedir_memory(int extra_pages_needed)
 {
 	int j, order, num_to_alloc = extra_pages_needed - extra_pages_allocated;
-	unsigned long flags = TOI_ATOMIC_GFP;
+	gfp_t flags = TOI_ATOMIC_GFP;
 
 	if (num_to_alloc < 1)
 		return 0;
@@ -300,7 +323,7 @@ static int toi_allocate_extra_pagedir_memory(int extra_pages_needed)
 		}
 
 		if (!virt) {
-			toi_kfree(7, extras_entry);
+			toi_kfree(7, extras_entry, sizeof(*extras_entry));
 			return extra_pages_allocated;
 		}
 
@@ -308,10 +331,7 @@ static int toi_allocate_extra_pagedir_memory(int extra_pages_needed)
 
 		extras_entry->page = newpage;
 		extras_entry->order = order;
-		extras_entry->next = NULL;
-
-		if (extras_list)
-			extras_entry->next = extras_list;
+		extras_entry->next = extras_list;
 
 		extras_list = extras_entry;
 
@@ -331,34 +351,28 @@ static int toi_allocate_extra_pagedir_memory(int extra_pages_needed)
  * real_nr_free_pages: Count pcp pages for a zone type or all zones
  * (-1 for all, otherwise zone_idx() result desired).
  */
-int real_nr_free_pages(unsigned long zone_idx_mask)
+unsigned long real_nr_free_pages(unsigned long zone_idx_mask)
 {
 	struct zone *zone;
-	int result = 0, i = 0, cpu;
+	int result = 0, cpu;
 
 	/* PCP lists */
-	for_each_zone(zone) {
-		if (!populated_zone(zone))
-			continue;
-
+	for_each_populated_zone(zone) {
 		if (!(zone_idx_mask & (1 << zone_idx(zone))))
 			continue;
 
 		for_each_online_cpu(cpu) {
-			struct per_cpu_pageset *pset = zone_pcp(zone, cpu);
-
-			for (i = 0; i < ARRAY_SIZE(pset->pcp); i++) {
-				struct per_cpu_pages *pcp;
-
-				pcp = &pset->pcp[i];
-				result += pcp->count;
-			}
+			struct per_cpu_pageset *pset =
+				per_cpu_ptr(zone->pageset, cpu);
+			struct per_cpu_pages *pcp = &pset->pcp;
+			result += pcp->count;
 		}
 
 		result += zone_page_state(zone, NR_FREE_PAGES);
 	}
 	return result;
 }
+EXPORT_SYMBOL_GPL(real_nr_free_pages);
 
 /*
  * Discover how much extra memory will be required by the drivers
@@ -367,25 +381,19 @@ int real_nr_free_pages(unsigned long zone_idx_mask)
  */
 static void get_extra_pd1_allowance(void)
 {
-	int orig_num_free = real_nr_free_pages(all_zones_mask), final;
+	unsigned long orig_num_free = real_nr_free_pages(all_zones_mask), final;
 
 	toi_prepare_status(CLEAR_BAR, "Finding allowance for drivers.");
 
-	suspend_console();
-	device_suspend(PMSG_FREEZE);
-	local_irq_disable(); /* irqs might have been re-enabled on us */
-	device_power_down(PMSG_FREEZE);
+	if (toi_go_atomic(PMSG_FREEZE, 1))
+		return;
 
 	final = real_nr_free_pages(all_zones_mask);
+	toi_end_atomic(ATOMIC_ALL_STEPS, 1, 0);
 
-	device_power_up();
-	local_irq_enable();
-	device_resume();
-	resume_console();
-
-	extra_pd1_pages_allowance = max(
-		orig_num_free - final + MIN_EXTRA_PAGES_ALLOWANCE,
-		MIN_EXTRA_PAGES_ALLOWANCE);
+	extra_pd1_pages_allowance = (orig_num_free > final) ?
+		orig_num_free - final + MIN_EXTRA_PAGES_ALLOWANCE :
+		MIN_EXTRA_PAGES_ALLOWANCE;
 }
 
 /*
@@ -393,25 +401,27 @@ static void get_extra_pd1_allowance(void)
  * expected compression ratio and possibly also ignoring our
  * allowance for extra pages.
  */
-static int main_storage_needed(int use_ecr,
+static unsigned long main_storage_needed(int use_ecr,
 		int ignore_extra_pd1_allow)
 {
-	return ((pagedir1.size + pagedir2.size +
+	return (pagedir1.size + pagedir2.size +
 	  (ignore_extra_pd1_allow ? 0 : extra_pd1_pages_allowance)) *
-	 (use_ecr ? toi_expected_compression_ratio() : 100) / 100);
+	 (use_ecr ? toi_expected_compression_ratio() : 100) / 100;
 }
 
 /*
  * Storage needed for the image header, in bytes until the return.
  */
-static int header_storage_needed(void)
+unsigned long get_header_storage_needed(void)
 {
-	int bytes = (int) sizeof(struct toi_header) +
+	unsigned long bytes = sizeof(struct toi_header) +
 			toi_header_storage_for_modules() +
-			toi_pageflags_space_needed();
+			toi_pageflags_space_needed() +
+			fs_info_space_needed();
 
 	return DIV_ROUND_UP(bytes, PAGE_SIZE);
 }
+EXPORT_SYMBOL_GPL(get_header_storage_needed);
 
 /*
  * When freeing memory, pages from either pageset might be freed.
@@ -432,46 +442,76 @@ static int header_storage_needed(void)
  * => any_to_free function
  */
 
-static int highpages_ps1_to_free(void)
+static unsigned long lowpages_usable_for_highmem_copy(void)
 {
-	return max_t(int, 0, DIV_ROUND_UP(get_highmem_size(pagedir1) -
-		get_highmem_size(pagedir2), 2) - real_nr_free_high_pages());
+	unsigned long needed = get_lowmem_size(pagedir1) +
+			extra_pd1_pages_allowance + MIN_FREE_RAM +
+			toi_memory_for_modules(0),
+		available = get_lowmem_size(pagedir2) +
+			 real_nr_free_low_pages() + extra_pages_allocated;
+
+	return available > needed ? available - needed : 0;
 }
 
-static int lowpages_ps1_to_free(void)
+static unsigned long highpages_ps1_to_free(void)
 {
-	return max_t(int, 0, DIV_ROUND_UP(get_lowmem_size(pagedir1) +
-		extra_pd1_pages_allowance + MIN_FREE_RAM +
-		toi_memory_for_modules(0) - get_lowmem_size(pagedir2) -
-		real_nr_free_low_pages() - extra_pages_allocated, 2));
+	unsigned long need = get_highmem_size(pagedir1),
+		      available = get_highmem_size(pagedir2) +
+			      real_nr_free_high_pages() +
+			      lowpages_usable_for_highmem_copy();
+
+	return need > available ? DIV_ROUND_UP(need - available, 2) : 0;
 }
 
-static int current_image_size(void)
+static unsigned long lowpages_ps1_to_free(void)
 {
-	return pagedir1.size + pagedir2.size + header_space_allocated;
+	unsigned long needed = get_lowmem_size(pagedir1) +
+			extra_pd1_pages_allowance + MIN_FREE_RAM +
+			toi_memory_for_modules(0),
+		available = get_lowmem_size(pagedir2) +
+			 real_nr_free_low_pages() + extra_pages_allocated;
+
+	return needed > available ? DIV_ROUND_UP(needed - available, 2) : 0;
 }
 
-static int storage_still_required(void)
+static unsigned long current_image_size(void)
 {
-	return max_t(int, 0, main_storage_needed(1, 1) - storage_available);
+	return pagedir1.size + pagedir2.size + header_storage_needed;
 }
 
-static int ram_still_required(void)
+static unsigned long storage_still_required(void)
 {
-	return max_t(int, 0, MIN_FREE_RAM + toi_memory_for_modules(0) -
-		real_nr_free_low_pages() + 2 * extra_pd1_pages_allowance);
+	unsigned long needed = main_storage_needed(1, 1);
+	return needed > storage_limit ? needed - storage_limit : 0;
 }
 
-static int any_to_free(int use_image_size_limit)
+static unsigned long ram_still_required(void)
 {
-	int user_limit = (use_image_size_limit && image_size_limit > 0) ?
-		max_t(int, 0, current_image_size() - (image_size_limit << 8))
-		: 0;
+	unsigned long needed = MIN_FREE_RAM + toi_memory_for_modules(0) +
+		2 * extra_pd1_pages_allowance,
+		  available = real_nr_free_low_pages() + extra_pages_allocated;
+	return needed > available ? needed - available : 0;
+}
 
-	int storage_limit = storage_still_required(),
-	    ram_limit = ram_still_required();
+unsigned long any_to_free(int use_image_size_limit)
+{
+	int use_soft_limit = use_image_size_limit && image_size_limit > 0;
+	unsigned long current_size = current_image_size(),
+		      soft_limit = use_soft_limit ? (image_size_limit << 8) : 0,
+		      to_free = use_soft_limit ? (current_size > soft_limit ?
+				      current_size - soft_limit : 0) : 0,
+		      storage_limit = storage_still_required(),
+		      ram_limit = ram_still_required(),
+		      first_max = max(to_free, storage_limit);
 
-	return max(max(user_limit, storage_limit), ram_limit);
+	return max(first_max, ram_limit);
+}
+
+static int need_pageset2(void)
+{
+	return (real_nr_free_low_pages() + extra_pages_allocated -
+		2 * extra_pd1_pages_allowance - MIN_FREE_RAM -
+		 toi_memory_for_modules(0) - pagedir1.size) < pagedir2.size;
 }
 
 /* amount_needed
@@ -479,7 +519,7 @@ static int any_to_free(int use_image_size_limit)
  * Calculates the amount by which the image size needs to be reduced to meet
  * our constraints.
  */
-static int amount_needed(int use_image_size_limit)
+static unsigned long amount_needed(int use_image_size_limit)
 {
 	return max(highpages_ps1_to_free() + lowpages_ps1_to_free(),
 			any_to_free(use_image_size_limit));
@@ -488,33 +528,30 @@ static int amount_needed(int use_image_size_limit)
 static int image_not_ready(int use_image_size_limit)
 {
 	toi_message(TOI_EAT_MEMORY, TOI_LOW, 1,
-		"Amount still needed (%d) > 0:%d. Header: %d < %d: %d,"
-		" Storage allocd: %d < %d: %d.\n",
+		"Amount still needed (%lu) > 0:%u,"
+		" Storage allocd: %lu < %lu: %u.\n",
 			amount_needed(use_image_size_limit),
 			(amount_needed(use_image_size_limit) > 0),
-			header_space_allocated, header_storage_needed(),
-			header_space_allocated < header_storage_needed(),
 			main_storage_allocated,
 			main_storage_needed(1, 1),
 			main_storage_allocated < main_storage_needed(1, 1));
 
 	toi_cond_pause(0, NULL);
 
-	return ((amount_needed(use_image_size_limit) > 0) ||
-		header_space_allocated < header_storage_needed() ||
-		 main_storage_allocated < main_storage_needed(1, 1));
+	return (amount_needed(use_image_size_limit) > 0) ||
+		 main_storage_allocated < main_storage_needed(1, 1);
 }
 
 static void display_failure_reason(int tries_exceeded)
 {
-	int storage_required = storage_still_required(),
+	unsigned long storage_required = storage_still_required(),
 	    ram_required = ram_still_required(),
 	    high_ps1 = highpages_ps1_to_free(),
 	    low_ps1 = lowpages_ps1_to_free();
 
 	printk(KERN_INFO "Failed to prepare the image because...\n");
 
-	if (!storage_available) {
+	if (!storage_limit) {
 		printk(KERN_INFO "- You need some storage available to be "
 				"able to hibernate.\n");
 		return;
@@ -525,33 +562,29 @@ static void display_failure_reason(int tries_exceeded)
 				"reached without successfully preparing the "
 				"image.\n");
 
-	if (header_space_allocated < header_storage_needed()) {
-		printk(KERN_INFO "- Insufficient header storage allocated. "
-				"Need %d, have %d.\n", header_storage_needed(),
-				header_space_allocated);
-		set_abort_result(TOI_INSUFFICIENT_STORAGE);
-	}
-
 	if (storage_required) {
-		printk(KERN_INFO " - We need at least %d pages of storage "
-				"(ignoring the header), but only have %d.\n",
+		printk(KERN_INFO " - We need at least %lu pages of storage "
+				"(ignoring the header), but only have %lu.\n",
 				main_storage_needed(1, 1),
 				main_storage_allocated);
 		set_abort_result(TOI_INSUFFICIENT_STORAGE);
 	}
 
 	if (ram_required) {
-		printk(KERN_INFO " - We need %d more free pages of low "
+		printk(KERN_INFO " - We need %lu more free pages of low "
 				"memory.\n", ram_required);
 		printk(KERN_INFO "     Minimum free     : %8d\n", MIN_FREE_RAM);
-		printk(KERN_INFO "   + Reqd. by modules : %8d\n",
+		printk(KERN_INFO "   + Reqd. by modules : %8lu\n",
 				toi_memory_for_modules(0));
-		printk(KERN_INFO "   - Currently free   : %8d\n",
-				real_nr_free_low_pages());
-		printk(KERN_INFO "   + 2 * extra allow  : %8d\n",
+		printk(KERN_INFO "   + 2 * extra allow  : %8lu\n",
 				2 * extra_pd1_pages_allowance);
+		printk(KERN_INFO "   - Currently free   : %8lu\n",
+				real_nr_free_low_pages());
+		printk(KERN_INFO "   - Pages allocd     : %8lu\n",
+				extra_pages_allocated);
 		printk(KERN_INFO "                      : ========\n");
-		printk(KERN_INFO "     Still needed     : %8d\n", ram_required);
+		printk(KERN_INFO "     Still needed     : %8lu\n",
+				ram_required);
 
 		/* Print breakdown of memory needed for modules */
 		toi_memory_for_modules(1);
@@ -559,13 +592,13 @@ static void display_failure_reason(int tries_exceeded)
 	}
 
 	if (high_ps1) {
-		printk(KERN_INFO "- We need to free %d highmem pageset 1 "
+		printk(KERN_INFO "- We need to free %lu highmem pageset 1 "
 				"pages.\n", high_ps1);
 		set_abort_result(TOI_UNABLE_TO_FREE_ENOUGH_MEMORY);
 	}
 
 	if (low_ps1) {
-		printk(KERN_INFO " - We need to free %d lowmem pageset 1 "
+		printk(KERN_INFO " - We need to free %ld lowmem pageset 1 "
 				"pages.\n", low_ps1);
 		set_abort_result(TOI_UNABLE_TO_FREE_ENOUGH_MEMORY);
 	}
@@ -575,8 +608,9 @@ static void display_stats(int always, int sub_extra_pd1_allow)
 {
 	char buffer[255];
 	snprintf(buffer, 254,
-		"Free:%d(%d). Sets:%d(%d),%d(%d). Header:%d/%d. Nosave:%d-%d"
-		"=%d. Storage:%u/%u(%u=>%u). Needed:%d,%d,%d(%d,%d,%d,%d)\n",
+		"Free:%lu(%lu). Sets:%lu(%lu),%lu(%lu). "
+		"Nosave:%lu-%lu=%lu. Storage:%lu/%lu(%lu=>%lu). "
+		"Needed:%lu,%lu,%lu(%u,%lu,%lu,%ld) (PS2:%s)\n",
 
 		/* Free */
 		real_nr_free_pages(all_zones_mask),
@@ -586,16 +620,13 @@ static void display_stats(int always, int sub_extra_pd1_allow)
 		pagedir1.size, pagedir1.size - get_highmem_size(pagedir1),
 		pagedir2.size, pagedir2.size - get_highmem_size(pagedir2),
 
-		/* Header */
-		header_space_allocated, header_storage_needed(),
-
 		/* Nosave */
 		num_nosave, extra_pages_allocated,
 		num_nosave - extra_pages_allocated,
 
 		/* Storage */
 		main_storage_allocated,
-		storage_available,
+		storage_limit,
 		main_storage_needed(1, sub_extra_pd1_allow),
 		main_storage_needed(1, 1),
 
@@ -603,10 +634,13 @@ static void display_stats(int always, int sub_extra_pd1_allow)
 		lowpages_ps1_to_free(), highpages_ps1_to_free(),
 		any_to_free(1),
 		MIN_FREE_RAM, toi_memory_for_modules(0),
-		extra_pd1_pages_allowance, image_size_limit << 8);
+		extra_pd1_pages_allowance,
+		image_size_limit,
+
+		need_pageset2() ? "yes" : "no");
 
 	if (always)
-		printk(buffer);
+		printk("%s", buffer);
 	else
 		toi_message(TOI_EAT_MEMORY, TOI_MEDIUM, 1, buffer);
 }
@@ -620,41 +654,55 @@ static void display_stats(int always, int sub_extra_pd1_allow)
  */
 static void generate_free_page_map(void)
 {
-	int order, loop, cpu;
-	struct page *page;
+	int order, cpu, t;
 	unsigned long flags, i;
 	struct zone *zone;
+	struct list_head *curr;
+	unsigned long pfn;
+	struct page *page;
 
-	for_each_zone(zone) {
-		if (!populated_zone(zone))
+	for_each_populated_zone(zone) {
+
+		if (!zone->spanned_pages)
 			continue;
 
 		spin_lock_irqsave(&zone->lock, flags);
 
-		for(i = 0; i < zone->spanned_pages; i++)
-			ClearPageNosaveFree(pfn_to_page(
-						zone->zone_start_pfn + i));
+		for (i = 0; i < zone->spanned_pages; i++) {
+			pfn = ZONE_START(zone) + i;
 
-		for (order = MAX_ORDER - 1; order >= 0; --order)
-			list_for_each_entry(page,
-					&zone->free_area[order].free_list, lru)
-				for(loop=0; loop < (1 << order); loop++)
-					SetPageNosaveFree(page+loop);
+			if (!pfn_valid(pfn))
+				continue;
 
-		
-		for_each_online_cpu(cpu) {
-			struct per_cpu_pageset *pset = zone_pcp(zone, cpu);
+			page = pfn_to_page(pfn);
 
-			for (i = 0; i < ARRAY_SIZE(pset->pcp); i++) {
-				struct per_cpu_pages *pcp;
-				struct page *page;
+			ClearPageNosaveFree(page);
+		}
 
-				pcp = &pset->pcp[i];
-				list_for_each_entry(page, &pcp->list, lru)
-					SetPageNosaveFree(page);
+		for_each_migratetype_order(order, t) {
+			list_for_each(curr,
+					&zone->free_area[order].free_list[t]) {
+				unsigned long j;
+
+				pfn = page_to_pfn(list_entry(curr, struct page,
+							lru));
+				for (j = 0; j < (1UL << order); j++)
+					SetPageNosaveFree(pfn_to_page(pfn + j));
 			}
 		}
 
+		for_each_online_cpu(cpu) {
+			struct per_cpu_pageset *pset =
+				per_cpu_ptr(zone->pageset, cpu);
+			struct per_cpu_pages *pcp = &pset->pcp;
+			struct page *page;
+			int t;
+
+			for (t = 0; t < MIGRATE_PCPTYPES; t++)
+				list_for_each_entry(page, &pcp->lists[t], lru)
+					SetPageNosaveFree(page);
+		}
+
 		spin_unlock_irqrestore(&zone->lock, flags);
 	}
 }
@@ -664,15 +712,15 @@ static void generate_free_page_map(void)
  * Description:	Return the number of pages that are free, beginning with and
  * 		including this one.
  */
-static int size_of_free_region(struct page *page)
+static int size_of_free_region(struct zone *zone, unsigned long start_pfn)
 {
-	struct zone *zone = page_zone(page);
-	struct page *posn = page, *last_in_zone =
-		pfn_to_page(zone->zone_start_pfn) + zone->spanned_pages - 1;
+	unsigned long this_pfn = start_pfn,
+		      end_pfn = ZONE_START(zone) + zone->spanned_pages - 1;
 
-	while (posn <= last_in_zone && PageNosaveFree(posn))
-		posn++;
-	return (posn - page);
+	while (this_pfn <= end_pfn && PageNosaveFree(pfn_to_page(this_pfn)))
+		this_pfn++;
+
+	return this_pfn - start_pfn;
 }
 
 /* flag_image_pages
@@ -696,7 +744,7 @@ static void flag_image_pages(int atomic_copy)
 
 	num_nosave = 0;
 
-	clear_dyn_pageflags(&pageset1_map);
+	memory_bm_clear(pageset1_map);
 
 	generate_free_page_map();
 
@@ -704,35 +752,35 @@ static void flag_image_pages(int atomic_copy)
 	 * Pages not to be saved are marked Nosave irrespective of being
 	 * reserved.
 	 */
-	for_each_zone(zone) {
+	for_each_populated_zone(zone) {
 		int highmem = is_highmem(zone);
 
-		if (!populated_zone(zone))
-			continue;
-
 		for (loop = 0; loop < zone->spanned_pages; loop++) {
-			unsigned long pfn = zone->zone_start_pfn + loop;
+			unsigned long pfn = ZONE_START(zone) + loop;
 			struct page *page;
 			int chunk_size;
 
 			if (!pfn_valid(pfn))
 				continue;
 
-			page = pfn_to_page(pfn);
-
-			chunk_size = size_of_free_region(page);
+			chunk_size = size_of_free_region(zone, pfn);
 			if (chunk_size) {
 				num_free += chunk_size;
 				loop += chunk_size - 1;
 				continue;
 			}
 
-			if (highmem)
-				page = saveable_highmem_page(pfn);
-			else
-				page = saveable_page(pfn);
+			page = pfn_to_page(pfn);
 
-			if (!page || PageNosave(page)) {
+			if (PageNosave(page)) {
+				num_nosave++;
+				continue;
+			}
+
+			page = highmem ? saveable_highmem_page(zone, pfn) :
+				saveable_page(zone, pfn);
+
+			if (!page) {
 				num_nosave++;
 				continue;
 			}
@@ -759,22 +807,23 @@ static void flag_image_pages(int atomic_copy)
 		}
 	}
 
-	if (atomic_copy)
-		return;
-
-	toi_message(TOI_EAT_MEMORY, TOI_MEDIUM, 0,
-		"Count data pages: Set1 (%d) + Set2 (%d) + Nosave (%d) + "
-		"NumFree (%d) = %d.\n",
-		pagedir1.size, pagedir2.size, num_nosave, num_free,
-		pagedir1.size + pagedir2.size + num_nosave + num_free);
+	if (!atomic_copy)
+		toi_message(TOI_EAT_MEMORY, TOI_MEDIUM, 0,
+			"Count data pages: Set1 (%d) + Set2 (%d) + Nosave (%ld)"
+						" + NumFree (%d) = %d.\n",
+			pagedir1.size, pagedir2.size, num_nosave, num_free,
+			pagedir1.size + pagedir2.size + num_nosave + num_free);
 }
 
 void toi_recalculate_image_contents(int atomic_copy)
 {
-	clear_dyn_pageflags(&pageset1_map);
+	memory_bm_clear(pageset1_map);
 	if (!atomic_copy) {
-		int pfn;
-		BITMAP_FOR_EACH_SET(&pageset2_map, pfn)
+		unsigned long pfn;
+		memory_bm_position_reset(pageset2_map);
+		for (pfn = memory_bm_next_pfn(pageset2_map);
+				pfn != BM_END_OF_MAP;
+				pfn = memory_bm_next_pfn(pageset2_map))
 			ClearPagePageset1Copy(pfn_to_page(pfn));
 		/* Need to call this before getting pageset1_size! */
 		toi_mark_pages_for_pageset2();
@@ -782,33 +831,42 @@ void toi_recalculate_image_contents(int atomic_copy)
 	flag_image_pages(atomic_copy);
 
 	if (!atomic_copy) {
-		storage_available = toiActiveAllocator->storage_available();
+		storage_limit = toiActiveAllocator->storage_available();
 		display_stats(0, 0);
 	}
 }
 
+int try_allocate_extra_memory(void)
+{
+	unsigned long wanted = pagedir1.size +  extra_pd1_pages_allowance -
+		get_lowmem_size(pagedir2);
+	if (wanted > extra_pages_allocated) {
+		unsigned long got = toi_allocate_extra_pagedir_memory(wanted);
+		if (wanted < got) {
+			toi_message(TOI_EAT_MEMORY, TOI_LOW, 1,
+				"Want %d extra pages for pageset1, got %d.\n",
+				wanted, got);
+			return 1;
+		}
+	}
+	return 0;
+}
+
+
 /* update_image
  *
  * Allocate [more] memory and storage for the image.
  */
-static void update_image(void)
+static void update_image(int ps2_recalc)
 {
-	int result, param_used, wanted, got;
+	int old_header_req;
+	unsigned long seek;
 
-	toi_recalculate_image_contents(0);
+	if (try_allocate_extra_memory())
+		return;
 
-	/* Include allowance for growth in pagedir1 while writing pagedir 2 */
-	wanted = pagedir1.size +  extra_pd1_pages_allowance -
-		get_lowmem_size(pagedir2);
-	if (wanted > extra_pages_allocated) {
-		got = toi_allocate_extra_pagedir_memory(wanted);
-		if (wanted < got) {
-			toi_message(TOI_EAT_MEMORY, TOI_LOW, 1,
-				"Want %d extra pages for pageset1, got %d.\n",
-				wanted, got);
-			return;
-		}
-	}
+	if (ps2_recalc)
+		goto recalc;
 
 	thaw_kernel_threads();
 
@@ -823,24 +881,32 @@ static void update_image(void)
 	 * don't complain if we can't get the full amount we're after.
 	 */
 
-	toiActiveAllocator->allocate_storage(
-		min(storage_available, main_storage_needed(0, 0)));
+	do {
+		int result;
 
-	main_storage_allocated = toiActiveAllocator->storage_allocated();
+		old_header_req = header_storage_needed;
+		toiActiveAllocator->reserve_header_space(header_storage_needed);
 
-	param_used = header_storage_needed();
+		/* How much storage is free with the reservation applied? */
+		storage_limit = toiActiveAllocator->storage_available();
+		seek = min(storage_limit, main_storage_needed(0, 0));
 
-	result = toiActiveAllocator->allocate_header_space(param_used);
+		result = toiActiveAllocator->allocate_storage(seek);
+		if (result)
+			printk("Failed to allocate storage (%d).\n", result);
 
-	if (result)
-		toi_message(TOI_EAT_MEMORY, TOI_LOW, 1,
-			"Still need to get more storage space for header.\n");
-	else
-		header_space_allocated = param_used;
+		main_storage_allocated =
+			toiActiveAllocator->storage_allocated();
+
+		/* Need more header because more storage allocated? */
+		header_storage_needed = get_header_storage_needed();
+
+	} while (header_storage_needed > old_header_req);
 
 	if (freeze_processes())
 		set_abort_result(TOI_FREEZING_FAILED);
 
+recalc:
 	toi_recalculate_image_contents(0);
 }
 
@@ -880,7 +946,7 @@ static int attempt_to_freeze(void)
  */
 static void eat_memory(void)
 {
-	int amount_wanted = 0;
+	unsigned long amount_wanted = 0;
 	int did_eat_memory = 0;
 
 	/*
@@ -894,7 +960,6 @@ static void eat_memory(void)
 	 * to freeze fail, we give up and abort.
 	 */
 
-	toi_recalculate_image_contents(0);
 	amount_wanted = amount_needed(1);
 
 	switch (image_size_limit) {
@@ -908,7 +973,6 @@ static void eat_memory(void)
 		drop_pagecache();
 		toi_recalculate_image_contents(0);
 		amount_wanted = amount_needed(1);
-		did_eat_memory = 1;
 		break;
 	default:
 		break;
@@ -916,44 +980,38 @@ static void eat_memory(void)
 
 	if (amount_wanted > 0 && !test_result_state(TOI_ABORTED) &&
 			image_size_limit != -1) {
-		struct zone *zone;
-		int zone_idx;
+		unsigned long request = amount_wanted;
+		unsigned long high_req = max(highpages_ps1_to_free(),
+				any_to_free(1));
+		unsigned long low_req = lowpages_ps1_to_free();
+		unsigned long got = 0;
 
 		toi_prepare_status(CLEAR_BAR,
-				"Seeking to free %dMB of memory.",
+				"Seeking to free %ldMB of memory.",
 				MB(amount_wanted));
 
 		thaw_kernel_threads();
 
-		for (zone_idx = 0; zone_idx < MAX_NR_ZONES; zone_idx++) {
-			unsigned long zone_type_free = max_t(int,
-					(zone_idx == ZONE_HIGHMEM) ?
-					highpages_ps1_to_free() :
-					lowpages_ps1_to_free(), amount_wanted);
+		/*
+		 * Ask for too many because shrink_all_memory doesn't
+		 * currently return enough most of the time.
+		 */
+		
+		if (low_req)
+			got = shrink_memory_mask(low_req, GFP_KERNEL);
+		if (high_req)
+			shrink_memory_mask(high_req - got, GFP_HIGHUSER);
 
-			if (zone_type_free < 0)
-				break;
+		did_eat_memory = 1;
 
-			for_each_zone(zone) {
-				if (zone_idx(zone) != zone_idx)
-					continue;
+		toi_recalculate_image_contents(0);
 
-				shrink_one_zone(zone, zone_type_free, 3);
+		amount_wanted = amount_needed(1);
 
-				did_eat_memory = 1;
-
-				toi_recalculate_image_contents(0);
-
-				amount_wanted = amount_needed(1);
-				zone_type_free = max_t(int,
-					(zone_idx == ZONE_HIGHMEM) ?
-					highpages_ps1_to_free() :
-					lowpages_ps1_to_free(), amount_wanted);
-
-				if (zone_type_free < 0)
-					break;
-			}
-		}
+		printk(KERN_DEBUG "Asked shrink_all_memory for %ld low pages &"
+				" %ld pages from anywhere, got %ld.\n",
+				high_req, low_req,
+				request - amount_wanted);
 
 		toi_cond_pause(0, NULL);
 
@@ -961,15 +1019,8 @@ static void eat_memory(void)
 			set_abort_result(TOI_FREEZING_FAILED);
 	}
 
-	if (did_eat_memory) {
-		unsigned long orig_state = get_toi_state();
-		/* Freeze_processes will call sys_sync too */
-		restore_toi_state(orig_state);
+	if (did_eat_memory)
 		toi_recalculate_image_contents(0);
-	}
-
-	/* Blank out image size display */
-	toi_update_status(100, 100, NULL);
 }
 
 /* toi_prepare_image
@@ -992,8 +1043,8 @@ int toi_prepare_image(void)
 {
 	int result = 1, tries = 1;
 
-	header_space_allocated = 0;
 	main_storage_allocated = 0;
+	no_ps2_needed = 0;
 
 	if (attempt_to_freeze())
 		return 1;
@@ -1001,9 +1052,11 @@ int toi_prepare_image(void)
 	if (!extra_pd1_pages_allowance)
 		get_extra_pd1_allowance();
 
-	storage_available = toiActiveAllocator->storage_available();
+	storage_limit = toiActiveAllocator->storage_available();
 
-	if (!storage_available) {
+	if (!storage_limit) {
+		printk(KERN_INFO "No storage available. Didn't try to prepare "
+				"an image.\n");
 		display_failure_reason(0);
 		set_abort_result(TOI_NOSTORAGE_AVAILABLE);
 		return 1;
@@ -1015,6 +1068,8 @@ int toi_prepare_image(void)
 		return 1;
 	}
 
+	toi_recalculate_image_contents(0);
+
 	do {
 		toi_prepare_status(CLEAR_BAR,
 				"Preparing Image. Try %d.", tries);
@@ -1024,7 +1079,7 @@ int toi_prepare_image(void)
 		if (test_result_state(TOI_ABORTED))
 			break;
 
-		update_image();
+		update_image(0);
 
 		tries++;
 
@@ -1040,14 +1095,17 @@ int toi_prepare_image(void)
 			abort_hibernate(TOI_UNABLE_TO_PREPARE_IMAGE,
 				"Unable to successfully prepare the image.\n");
 		} else {
-			unlink_lru_lists();
+			/* Pageset 2 needed? */
+			if (!need_pageset2() &&
+				  test_action_state(TOI_NO_PS2_IF_UNNEEDED)) {
+				no_ps2_needed = 1;
+				toi_recalculate_image_contents(0);
+				update_image(1);
+			}
+
 			toi_cond_pause(1, "Image preparation complete.");
 		}
 	}
 
 	return result ? result : allocate_checksum_pages();
 }
-
-#ifdef CONFIG_TOI_EXPORTS
-EXPORT_SYMBOL_GPL(real_nr_free_pages);
-#endif
