@@ -1,7 +1,7 @@
 /*
  * kernel/power/tuxonice_prepare_image.h
  *
- * Copyright (C) 2003-2007 Nigel Cunningham (nigel at tuxonice net)
+ * Copyright (C) 2003-2010 Nigel Cunningham (nigel at tuxonice net)
  *
  * This file is released under the GPLv2.
  *
@@ -11,10 +11,10 @@
 
 extern int toi_prepare_image(void);
 extern void toi_recalculate_image_contents(int storage_available);
-extern int real_nr_free_pages(unsigned long zone_idx_mask);
-extern int image_size_limit;
+extern unsigned long real_nr_free_pages(unsigned long zone_idx_mask);
+extern long image_size_limit;
 extern void toi_free_extra_pagedir_memory(void);
-extern int extra_pd1_pages_allowance;
+extern unsigned long extra_pd1_pages_allowance;
 extern void free_attention_list(void);
 
 #define MIN_FREE_RAM 100
@@ -33,3 +33,6 @@ extern void free_attention_list(void);
 #define ZONE_HIGHMEM (MAX_NR_ZONES + 1)
 #endif
 
+unsigned long get_header_storage_needed(void);
+unsigned long any_to_free(int use_image_size_limit);
+int try_allocate_extra_memory(void);
