@@ -1,7 +1,7 @@
 /*
  * kernel/power/tuxonice_storage.c
  *
- * Copyright (C) 2005-2007 Nigel Cunningham (nigel at tuxonice net)
+ * Copyright (C) 2005-2010 Nigel Cunningham (nigel at tuxonice net)
  *
  * This file is released under the GPLv2.
  *
@@ -154,14 +154,14 @@ static void storage_manager_simulate(void)
 
 static int usm_storage_needed(void)
 {
-	return strlen(usm_helper_data.program);
+	return sizeof(int) + strlen(usm_helper_data.program) + 1;
 }
 
 static int usm_save_config_info(char *buf)
 {
 	int len = strlen(usm_helper_data.program);
-	memcpy(buf, usm_helper_data.program, len);
-	return len;
+	memcpy(buf, usm_helper_data.program, len + 1);
+	return sizeof(int) + len + 1;
 }
 
 static void usm_load_config_info(char *buf, int size)
@@ -170,13 +170,13 @@ static void usm_load_config_info(char *buf, int size)
 	if (usm_helper_data.program[0])
 		return;
 
-	memcpy(usm_helper_data.program, buf, size);
+	memcpy(usm_helper_data.program, buf + sizeof(int), *((int *) buf));
 }
 
 static int usm_memory_needed(void)
 {
 	/* ball park figure of 32 pages */
-	return (32 * PAGE_SIZE);
+	return 32 * PAGE_SIZE;
 }
 
 /* toi_prepare_usm
@@ -201,7 +201,7 @@ int toi_prepare_usm(void)
 
 	toi_activate_storage(0);
 
-	return (usm_helper_data.pid != -1);
+	return usm_helper_data.pid != -1;
 }
 
 void toi_cleanup_usm(void)
@@ -232,23 +232,12 @@ static void storage_manager_activate(void)
  */
 
 static struct toi_sysfs_data sysfs_params[] = {
-	{ TOI_ATTR("simulate_atomic_copy", SYSFS_RW),
-	  .type				= TOI_SYSFS_DATA_NONE,
-	  .write_side_effect		= storage_manager_simulate,
-	},
-
-	{ TOI_ATTR("enabled", SYSFS_RW),
-	  SYSFS_INT(&usm_ops.enabled, 0, 1, 0)
-	},
-
-	{ TOI_ATTR("program", SYSFS_RW),
-	  SYSFS_STRING(usm_helper_data.program, 254, 0)
-	},
-
-	{ TOI_ATTR("activate_storage", SYSFS_RW),
-	  SYSFS_INT(&storage_manager_action, 0, 1, 0),
-	  .write_side_effect		= storage_manager_activate,
-	}
+	SYSFS_NONE("simulate_atomic_copy", storage_manager_simulate),
+	SYSFS_INT("enabled", SYSFS_RW, &usm_ops.enabled, 0, 1, 0, NULL),
+	SYSFS_STRING("program", SYSFS_RW, usm_helper_data.program, 254, 0,
+		NULL),
+	SYSFS_INT("activate_storage", SYSFS_RW , &storage_manager_action, 0, 1,
+			0, storage_manager_activate)
 };
 
 static struct toi_module_ops usm_ops = {
@@ -279,7 +268,7 @@ int toi_usm_init(void)
 	usm_helper_data.netlink_id = NETLINK_TOI_USM;
 	usm_helper_data.name = "userspace storage manager";
 	usm_helper_data.rcv_msg = usm_user_rcv_msg;
-	usm_helper_data.interface_version = 1;
+	usm_helper_data.interface_version = 2;
 	usm_helper_data.must_init = 0;
 	init_completion(&usm_helper_data.wait_for_process);
 
@@ -288,5 +277,6 @@ int toi_usm_init(void)
 
 void toi_usm_exit(void)
 {
+	toi_netlink_close_complete(&usm_helper_data);
 	toi_unregister_module(&usm_ops);
 }
